#
# Copyright (c) 2021 Nitric Technologies Pty Ltd.
#
# This file is part of Nitric Python 3 SDK.
# See https://github.com/nitrictech/python-sdk for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: nitric/proto/queues/v1/queues.proto
# plugin: python-betterproto
# This file has been @generated

from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Dict,
    List,
    Optional,
)

import betterproto
import betterproto.lib.google.protobuf as betterproto_lib_google_protobuf
import grpclib
from betterproto.grpc.grpclib_server import ServiceBase


if TYPE_CHECKING:
    import grpclib.server
    from betterproto.grpc.grpclib_client import MetadataLike
    from grpclib.metadata import Deadline


@dataclass(eq=False, repr=False)
class QueueEnqueueRequest(betterproto.Message):
    queue_name: str = betterproto.string_field(1)
    """
    The Nitric name for the queue this will automatically be resolved to the
    provider specific queue identifier.
    """

    messages: List["QueueMessage"] = betterproto.message_field(2)
    """Array of messages to push to the queue"""


@dataclass(eq=False, repr=False)
class QueueEnqueueResponse(betterproto.Message):
    """Response for sending messages to a queue"""

    failed_messages: List["FailedEnqueueMessage"] = betterproto.message_field(1)
    """A list of messages that failed to be queued"""


@dataclass(eq=False, repr=False)
class QueueDequeueRequest(betterproto.Message):
    queue_name: str = betterproto.string_field(1)
    """
    The nitric name for the queue this will automatically be resolved to the
    provider specific queue identifier.
    """

    depth: int = betterproto.int32_field(2)
    """
    The max number of messages to pop off the queue, may be capped by provider
    specific limitations
    """


@dataclass(eq=False, repr=False)
class QueueDequeueResponse(betterproto.Message):
    messages: List["DequeuedMessage"] = betterproto.message_field(1)
    """Array of messages popped off the queue"""


@dataclass(eq=False, repr=False)
class QueueCompleteRequest(betterproto.Message):
    queue_name: str = betterproto.string_field(1)
    """
    The nitric name for the queue  this will automatically be resolved to the
    provider specific queue identifier.
    """

    lease_id: str = betterproto.string_field(2)
    """Lease id of the message to be completed"""


@dataclass(eq=False, repr=False)
class QueueCompleteResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class QueueMessage(betterproto.Message):
    """An message to be sent to a queue."""

    struct_payload: "betterproto_lib_google_protobuf.Struct" = (
        betterproto.message_field(1, group="content")
    )


@dataclass(eq=False, repr=False)
class DequeuedMessage(betterproto.Message):
    lease_id: str = betterproto.string_field(1)
    message: "QueueMessage" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class FailedEnqueueMessage(betterproto.Message):
    message: "QueueMessage" = betterproto.message_field(1)
    """The message that failed to be pushed"""

    details: str = betterproto.string_field(2)
    """A description of the failure"""


class QueuesStub(betterproto.ServiceStub):
    async def enqueue(
        self,
        queue_enqueue_request: "QueueEnqueueRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "QueueEnqueueResponse":
        return await self._unary_unary(
            "/nitric.proto.queues.v1.Queues/Enqueue",
            queue_enqueue_request,
            QueueEnqueueResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def dequeue(
        self,
        queue_dequeue_request: "QueueDequeueRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "QueueDequeueResponse":
        return await self._unary_unary(
            "/nitric.proto.queues.v1.Queues/Dequeue",
            queue_dequeue_request,
            QueueDequeueResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def complete(
        self,
        queue_complete_request: "QueueCompleteRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "QueueCompleteResponse":
        return await self._unary_unary(
            "/nitric.proto.queues.v1.Queues/Complete",
            queue_complete_request,
            QueueCompleteResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )


class QueuesBase(ServiceBase):
    async def enqueue(
        self, queue_enqueue_request: "QueueEnqueueRequest"
    ) -> "QueueEnqueueResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def dequeue(
        self, queue_dequeue_request: "QueueDequeueRequest"
    ) -> "QueueDequeueResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def complete(
        self, queue_complete_request: "QueueCompleteRequest"
    ) -> "QueueCompleteResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_enqueue(
        self, stream: "grpclib.server.Stream[QueueEnqueueRequest, QueueEnqueueResponse]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.enqueue(request)
        await stream.send_message(response)

    async def __rpc_dequeue(
        self, stream: "grpclib.server.Stream[QueueDequeueRequest, QueueDequeueResponse]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.dequeue(request)
        await stream.send_message(response)

    async def __rpc_complete(
        self,
        stream: "grpclib.server.Stream[QueueCompleteRequest, QueueCompleteResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.complete(request)
        await stream.send_message(response)

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/nitric.proto.queues.v1.Queues/Enqueue": grpclib.const.Handler(
                self.__rpc_enqueue,
                grpclib.const.Cardinality.UNARY_UNARY,
                QueueEnqueueRequest,
                QueueEnqueueResponse,
            ),
            "/nitric.proto.queues.v1.Queues/Dequeue": grpclib.const.Handler(
                self.__rpc_dequeue,
                grpclib.const.Cardinality.UNARY_UNARY,
                QueueDequeueRequest,
                QueueDequeueResponse,
            ),
            "/nitric.proto.queues.v1.Queues/Complete": grpclib.const.Handler(
                self.__rpc_complete,
                grpclib.const.Cardinality.UNARY_UNARY,
                QueueCompleteRequest,
                QueueCompleteResponse,
            ),
        }



#
# Copyright (c) 2021 Nitric Technologies Pty Ltd.
#
# This file is part of Nitric Python 3 SDK.
# See https://github.com/nitrictech/python-sdk for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

import functools
import inspect
import json
from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Dict, Generic, List, Optional, Protocol, TypeVar, Union

from opentelemetry import propagate

from nitric.proto.schedules.v1 import ServerMessage as ScheduleServerMessage
from nitric.proto.topics.v1 import ClientMessage as TopicClientMessage
from nitric.proto.topics.v1 import MessageResponse as TopicResponse
from nitric.proto.topics.v1 import ServerMessage as TopicServerMessage
from nitric.utils import dict_from_struct

Record = Dict[str, Union[str, List[str]]]
PROPAGATOR = propagate.get_global_textmap()


class HttpMethod(Enum):
    """Valid query expression operators."""

    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"
    PATCH = "PATCH"
    OPTIONS = "OPTIONS"

    def __str__(self):
        return str(self.value)


class TriggerContext(Protocol):
    """Represents an abstract request/response context for any trigger."""

    @staticmethod
    def _from_request(server_message) -> TriggerContext:
        """Convert a server message request into this trigger context."""
        pass

    def to_response(self):
        """Convert this trigger context into a server message response."""
        pass


# ====== HTTP ======


class HttpRequest:
    """Represents a translated Http Request forwarded from the Nitric Membrane."""

    def __init__(
        self,
        data: bytes,
        method: str,
        path: str,
        params: Dict[str, str],
        query: Record,
        headers: Record,
    ):
        """Construct a new HttpRequest."""
        self.data = data
        self.method = method
        self.path = path
        self.params = params
        self.query = query
        self.headers = headers

    @property
    def json(self) -> Optional[Any]:
        """Get the body of the request as JSON, returns None if request body is not JSON."""
        try:
            return json.loads(self.body)
        except json.JSONDecodeError:
            return None
        except TypeError:
            return None

    @property
    def body(self):
        """Get the body of the request as text."""
        return self.data.decode("utf-8")


class HttpResponse:
    """Represents an HTTP Response to be generated by the Nitric Membrane in response to an HTTP Request Trigger."""

    def __init__(self, status: int = 200, headers: Optional[Record] = None, body: Optional[bytes] = None):
        """Construct a new HttpResponse."""
        self.status = status
        self.headers = headers if headers else {}
        self._body = body if body else bytes()

    @property
    def body(self):
        """Return the HTTP response body."""
        return self._body

    @body.setter
    def body(self, value: Union[str, bytes, Any]):
        if isinstance(value, str):
            self._body = value.encode("utf-8")
        elif isinstance(value, bytes):
            self._body = value
        else:
            self._body = json.dumps(value).encode("utf-8")
            self.headers["Content-Type"] = ["application/json"]


class HttpContext:
    """Represents the full request/response context for an Http based trigger."""

    def __init__(self, request: HttpRequest, response: Optional[HttpResponse] = None):
        """Construct a new HttpContext."""
        self.req = request
        self.res = response if response else HttpResponse()

    @staticmethod
    def _ensure_value_is_list(value: Union[str, List[str]]) -> List[str]:
        return list(value) if isinstance(value, list) else [value]


class MessageRequest:
    """Represents a translated Event, from a Subscribed Topic, forwarded from the Nitric Membrane."""

    data: dict[str, Any]
    topic: str

    def __init__(self, data: dict[str, Any], topic: str):
        """Construct a new EventRequest."""
        self.data = data
        self.topic = topic


class MessageResponse:
    """Represents the response to a trigger from an Event as a result of a Topic subscription."""

    def __init__(self, success: bool = True):
        """Construct a new EventResponse."""
        self.success = success


class MessageContext:
    """Represents the full request/response context for an Event based trigger."""

    def __init__(self, request: MessageRequest, response: Optional[MessageResponse] = None):
        """Construct a new EventContext."""
        self.req = request
        self.res = response if response else MessageResponse()

    @staticmethod
    def _from_request(msg: TopicServerMessage) -> MessageContext:
        """Construct a new EventContext from a Topic trigger from the Nitric Membrane."""
        return MessageContext(
            request=MessageRequest(
                data=dict_from_struct(msg.message_request.message.struct_payload),
                topic=msg.message_request.topic_name,
            )
        )

    def to_response(self) -> TopicClientMessage:
        """Construct a EventContext for the Nitric Membrane from this context object."""
        return TopicClientMessage(message_response=TopicResponse(success=self.res.success))


# == WEBSOCKET ==


class WebsocketRequest:
    """Represents an incoming websocket event."""

    def __init__(self, connection_id: str):
        """Construct a new WebsocketRequest."""
        self.connection_id = connection_id


class WebsocketConnectionRequest(WebsocketRequest):
    """Represents an incoming websocket connection."""

    query: Dict[str, str | List[str]]

    def __init__(self, connection_id: str, query: Dict[str, str | List[str]]):
        """Construct a new WebsocketConnectionRequest."""
        super().__init__(connection_id=connection_id)
        self.query = query


class WebsocketMessageRequest(WebsocketRequest):
    """Represents an incoming websocket message."""

    data: bytes

    def __init__(self, connection_id: str, data: bytes):
        """Construct a new WebsocketMessageRequest."""
        super().__init__(connection_id=connection_id)
        self.data = data


class WebsocketResponse:
    """Represents a response to a websocket event."""

    def __init__(self):
        """Construct a new WebsocketResponse."""


class WebsocketConnectionResponse(WebsocketResponse):
    """Represents a response to a websocket connection event."""

    reject: bool

    def __init__(self, reject: bool = False):
        """Construct a new WebsocketConnectionResponse."""
        self.reject = reject


AnyWebsocketRequest = Union[WebsocketRequest, WebsocketConnectionRequest, WebsocketMessageRequest]
AnyWebsocketResponse = Union[WebsocketResponse, WebsocketConnectionResponse]


class WebsocketContext:
    """Represents the full request/response context for a Websocket based trigger."""

    def __init__(self, request: AnyWebsocketRequest, response: Optional[AnyWebsocketResponse] = None):
        """Construct a new WebsocketContext."""
        self.req = request
        if response:
            self.res = response
        elif isinstance(request, WebsocketConnectionRequest):
            self.res = WebsocketConnectionResponse()
        else:
            self.res = WebsocketResponse()


# == Schedules ==


class IntervalRequest:
    """Represents a translated Event, from a Schedule, forwarded from the Nitric Membrane."""

    def __init__(self, schedule_name: str):
        """Construct a new IntervalRequest."""
        self.schedule_name = schedule_name


class IntervalResponse:
    """Represents the response to a trigger from an Interval as a result of a Schedule."""

    _request_id: str

    def __init__(self, request_id: str):
        """Construct a new IntervalResponse."""
        self._request_id = request_id


class IntervalContext:
    """Represents the full request/response context for a scheduled trigger."""

    def __init__(self, msg: ScheduleServerMessage):
        """Construct a new EventContext."""
        self.req = IntervalRequest(schedule_name=msg.interval_request.schedule_name)
        self.res = IntervalResponse(msg.id)


C = TypeVar("C")


class Middleware(Protocol, Generic[C]):
    """A middleware function."""

    async def __call__(self, ctx: C, nxt: Optional[Middleware[C]] = None) -> C:
        """Process trigger context."""
        ...


class Handler(Protocol, Generic[C]):
    """A handler function."""

    async def __call__(self, ctx: C) -> C | None:
        """Process trigger context."""
        ...


HttpMiddleware = Middleware[HttpContext]
EventMiddleware = Middleware[MessageContext]
IntervalMiddleware = Middleware[IntervalContext]
WebsocketMiddleware = Middleware[WebsocketContext]

HttpHandler = Handler[HttpContext]
EventHandler = Handler[MessageContext]
IntervalHandler = Handler[IntervalContext]
WebsocketHandler = Handler[WebsocketContext]


def _convert_to_middleware(handler: Handler[C] | Middleware[C]) -> Middleware[C]:
    """Convert a handler to a middleware, if it's already a middleware it's returned unchanged."""
    if not _is_handler(handler):
        # it's not a middleware, don't convert it.
        return handler  # type: ignore

    async def middleware(ctx: C, nxt: Middleware[C]) -> C:
        context = await handler(ctx)  # type: ignore
        return await nxt(context) if nxt else context  # type: ignore

    return middleware  # type: ignore


def _is_handler(unknown: Middleware[C] | Handler[C]) -> bool:
    """Return True if the provided function is a handler (1 positional arg)."""
    signature = inspect.signature(unknown)
    params = signature.parameters
    positional = [name for name, param in params.items() if param.default == inspect.Parameter.empty]
    return len(positional) == 1


def compose_middleware(*middlewares: Middleware[C] | Handler[C]) -> Middleware[C]:
    """
    Compose multiple middleware functions into a single middleware function.

    The resulting middleware will effectively be a chain of the provided middleware,
    where each calls the next in the chain when they're successful.
    """
    middlewares = [_convert_to_middleware(middleware) for middleware in middlewares]  # type: ignore

    async def composed(ctx: C, nxt: Optional[Middleware[C]] = None) -> C:
        last_middleware = nxt

        def reduce_chain(acc_next: Middleware[C], cur: Middleware[C]) -> Middleware[C]:
            async def chained_middleware(ctx: C, nxt: Optional[Middleware[C]] = None) -> C:
                result = (await nxt(ctx)) if nxt is not None else ctx  # type: ignore
                # type ignored because mypy appears to misidentify the correct return type
                output_context = await cur(result, acc_next)  # type: ignore
                if not output_context:
                    return result  # type: ignore
                return output_context  # type: ignore

            return chained_middleware

        middleware_chain = functools.reduce(reduce_chain, reversed(middlewares), last_middleware)  # type: ignore
        # type ignored because mypy appears to misidentify the correct return type
        return await middleware_chain(ctx)  # type: ignore

    return composed


class FunctionServer(ABC):
    """Represents a worker that should be started at runtime."""

    @abstractmethod
    async def start(self) -> None:
        """Start the worker."""



#
# Copyright (c) 2021 Nitric Technologies Pty Ltd.
#
# This file is part of Nitric Python 3 SDK.
# See https://github.com/nitrictech/python-sdk for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Callable, Concatenate, Dict, List, Optional, ParamSpec, TypeVar, Union

import betterproto
import grpclib
from grpclib import GRPCError

from nitric.application import Nitric
from nitric.bidi import AsyncNotifierList
from nitric.context import (
    FunctionServer,
    HttpContext,
    HttpHandler,
    HttpMethod,
    HttpMiddleware,
    HttpRequest,
    Record,
    compose_middleware,
)
from nitric.exception import exception_from_grpc_error
from nitric.proto.apis.v1 import (
    ApiDetailsRequest,
    ApiStub,
    ApiWorkerOptions,
    ApiWorkerScopes,
    ClientMessage,
    HeaderValue,
)
from nitric.proto.apis.v1 import HttpRequest as ProtoHttpRequest
from nitric.proto.apis.v1 import HttpResponse as ProtoHttpResponse
from nitric.proto.apis.v1 import RegistrationRequest
from nitric.proto.resources.v1 import (
    ApiOpenIdConnectionDefinition,
    ApiResource,
    ApiScopes,
    ApiSecurityDefinitionResource,
    ResourceDeclareRequest,
    ResourceIdentifier,
    ResourceType,
)
from nitric.resources.resource import Resource as BaseResource
from nitric.utils import new_default_channel


@dataclass
class ApiDetails:
    """
    Represents a deployed API's details.

    url (str): the url of the API
    """

    url: str


@dataclass
class JwtSecurityDefinition:
    """
    Represents the JWT security definition for an API.

    issuer (str): the JWT issuer
    audiences (List[str]): a list of the allowed audiences for the API
    """

    issuer: str
    audiences: List[str]


@dataclass
class MethodOptions:
    """
    Represents options when defining a method handler.

    security (dict[str, List[str]])
    """

    security: Optional[List[ScopedOidcOptions]] = None


# SecurityDefinition = JwtSecurityDefinition


class ApiOptions:
    """Represents options when creating an API, such as middleware to be applied to all HTTP request to the API."""

    path: str
    middleware: Optional[Union[HttpMiddleware, List[HttpMiddleware]]]
    security: Optional[List[ScopedOidcOptions]]

    def __init__(
        self,
        path: str = "",
        middleware: Optional[Union[HttpMiddleware, List[HttpMiddleware]]] = None,
        security: Optional[List[ScopedOidcOptions]] = None,
    ):
        """Construct a new API options object."""
        if middleware is None:
            middleware = []
        if security is None:
            security = []
        self.middleware = middleware
        self.security = security
        self.path = path


class RouteOptions:
    """Represents options when creating a route, such as middleware to be applied to all HTTP Methods for the route."""

    middleware: Union[None, List[HttpMiddleware]]

    def __init__(self, middleware: Optional[List[HttpMiddleware]] = None):
        """Construct a new route options object."""
        if middleware is None:
            middleware = []
        self.middleware = middleware


def _to_resource_identifier(b: Api) -> ResourceIdentifier:
    return ResourceIdentifier(name=b.name, type=ResourceType.Api)


def _security_to_grpc_declaration(security: Optional[List[ScopedOidcOptions]] = None) -> dict[str, ApiScopes]:
    if security is None or len(security) == 0:
        return {}
    return {sec.name: ApiScopes(sec.scopes) for sec in security}


Param = ParamSpec("Param")
RetType = TypeVar("RetType")
OriginalFunc = Callable[Param, RetType]
DecoratedFunc = Callable[Concatenate[str, Param], RetType]


class Api(BaseResource):
    """An HTTP API."""

    app: Nitric
    name: str
    path: str
    middleware: List[HttpMiddleware]
    routes: List[Route]
    security: Optional[List[ScopedOidcOptions]]
    _api_stub: ApiStub

    def __init__(self, name: str, opts: Optional[ApiOptions] = None):
        """Construct a new HTTP API."""
        super().__init__(name)
        self._api_stub = ApiStub(channel=self._channel)
        if opts is None:
            opts = ApiOptions()

        self.middleware = (
            opts.middleware
            if isinstance(opts.middleware, list)
            else [opts.middleware] if opts.middleware is not None else []
        )
        self.path = opts.path
        self.routes = []
        self.security = opts.security

    async def _register(self) -> None:
        try:
            await self._resources_stub.declare(
                resource_declare_request=ResourceDeclareRequest(
                    id=_to_resource_identifier(self),
                    api=ApiResource(
                        security=_security_to_grpc_declaration(self.security),
                    ),
                )
            )

            # Attach security rules for this api
            for security_rule in self.security if self.security else []:
                _attach_oidc(api_name=self.name, options=security_rule)
        except GRPCError as grpc_err:
            raise exception_from_grpc_error(grpc_err) from grpc_err

    def _route(self, match: str, opts: Optional[RouteOptions] = None) -> Route:
        """Define an HTTP route to be handled by this API."""
        if opts is None:
            opts = RouteOptions()

        if self.middleware is not None:
            opts.middleware = (self.middleware + opts.middleware) if opts.middleware is not None else self.middleware

        r = Route(self, match, opts)
        self.routes.append(r)
        return r

    def all(self, match: str, opts: Optional[MethodOptions] = None) -> Callable[[HttpHandler], None]:
        """Define an HTTP route which will respond to HTTP GET requests."""

        def decorator(function: HttpHandler) -> None:
            r = self._route(match)
            r.method(
                [
                    HttpMethod.GET,
                    HttpMethod.POST,
                    HttpMethod.PATCH,
                    HttpMethod.PUT,
                    HttpMethod.DELETE,
                    HttpMethod.OPTIONS,
                ],
                function,
                opts=opts if opts is not None else MethodOptions(security=None),
            )

        return decorator

    def methods(
        self, methods: List[HttpMethod], match: str, opts: Optional[MethodOptions] = None
    ) -> Callable[[HttpHandler], None]:
        """Define an HTTP route which will respond to specific HTTP requests defined by a list of verbs."""
        if opts is None:
            opts = MethodOptions()

        def decorator(function: HttpHandler) -> None:
            r = self._route(match)
            r.method(methods, function, opts=opts)

        return decorator

    def get(self, match: str, opts: Optional[MethodOptions] = None) -> Callable[[HttpHandler], None]:
        """Define an HTTP route which will respond to HTTP GET requests."""
        if opts is None:
            opts = MethodOptions()

        def decorator(function: HttpHandler) -> None:
            r = self._route(match)
            r.get(function, opts=opts)

        return decorator

    def post(self, match: str, opts: Optional[MethodOptions] = None) -> Callable[[HttpHandler], None]:
        """Define an HTTP route which will respond to HTTP POST requests."""
        if opts is None:
            opts = MethodOptions()

        def decorator(function: HttpHandler) -> None:
            r = self._route(match)
            r.post(function, opts=opts)

        return decorator

    def delete(self, match: str, opts: Optional[MethodOptions] = None) -> Callable[[HttpHandler], None]:
        """Define an HTTP route which will respond to HTTP DELETE requests."""
        if opts is None:
            opts = MethodOptions()

        def decorator(function: HttpHandler) -> None:
            r = self._route(match)
            r.delete(function, opts=opts)

        return decorator

    def options(self, match: str, opts: Optional[MethodOptions] = None) -> Callable[[HttpHandler], None]:
        """Define an HTTP route which will respond to HTTP OPTIONS requests."""
        if opts is None:
            opts = MethodOptions()

        def decorator(function: HttpHandler) -> None:
            r = self._route(match)
            r.options(function, opts=opts)

        return decorator

    def patch(self, match: str, opts: Optional[MethodOptions] = None) -> Callable[[HttpHandler], None]:
        """Define an HTTP route which will respond to HTTP PATCH requests."""
        if opts is None:
            opts = MethodOptions()

        def decorator(function: HttpHandler) -> None:
            r = self._route(match)
            r.patch(function, opts=opts)

        return decorator

    def put(self, match: str, opts: Optional[MethodOptions] = None) -> Callable[[HttpHandler], None]:
        """Define an HTTP route which will respond to HTTP PUT requests."""
        if opts is None:
            opts = MethodOptions()

        def decorator(function: HttpHandler) -> None:
            r = self._route(match)
            r.put(function, opts=opts)

        return decorator

    async def _details(self) -> ApiDetails:
        """Get the API deployment details."""
        try:
            res = await self._api_stub.api_details(
                api_details_request=ApiDetailsRequest(
                    api_name=self.name,
                )
            )
            return ApiDetails(res.url)
        except GRPCError as grpc_err:
            raise exception_from_grpc_error(grpc_err) from grpc_err

    async def url(self) -> str:
        """Get the APIs live URL."""
        details = await self._details()
        return details.url


class Route:
    """An HTTP route."""

    api: Api
    path: str
    middleware: List[HttpMiddleware]

    def __init__(self, api: Api, path: str, opts: RouteOptions):
        """Define a route to be handled by the provided API."""
        self.api = api
        self.path = (api.path + path).replace("//", "/")
        self.middleware = opts.middleware if opts.middleware is not None else []

    def method(
        self, methods: List[HttpMethod], *middleware: HttpMiddleware | HttpHandler, opts: Optional[MethodOptions] = None
    ) -> None:
        """Register middleware for multiple HTTP Methods."""

        # ensure route/api middlewares are added
        middleware = (
            *self.middleware,
            *middleware
        )

        Method(self, methods, *middleware, opts=opts if opts else MethodOptions())

    def get(self, *middleware: HttpMiddleware | HttpHandler, opts: Optional[MethodOptions] = None) -> None:
        """Register middleware for HTTP GET requests."""
        return self.method([HttpMethod.GET], *middleware, opts=opts)

    def post(self, *middleware: HttpMiddleware | HttpHandler, opts: Optional[MethodOptions] = None) -> None:
        """Register middleware for HTTP POST requests."""
        return self.method([HttpMethod.POST], *middleware, opts=opts)

    def put(self, *middleware: HttpMiddleware | HttpHandler, opts: Optional[MethodOptions] = None) -> None:
        """Register middleware for HTTP PUT requests."""
        return self.method([HttpMethod.PUT], *middleware, opts=opts)

    def patch(self, *middleware: HttpMiddleware | HttpHandler, opts: Optional[MethodOptions] = None) -> None:
        """Register middleware for HTTP PATCH requests."""
        return self.method([HttpMethod.PATCH], *middleware, opts=opts)

    def delete(self, *middleware: HttpMiddleware | HttpHandler, opts: Optional[MethodOptions] = None) -> None:
        """Register middleware for HTTP DELETE requests."""
        return self.method([HttpMethod.DELETE], *middleware, opts=opts)

    def options(self, *middleware: HttpMiddleware | HttpHandler, opts: Optional[MethodOptions] = None) -> None:
        """Register middleware for HTTP OPTIONS requests."""
        return self.method([HttpMethod.OPTIONS], *middleware, opts=opts)


class Method:
    """A method handler."""

    server: ApiRouteWorker
    route: Route
    methods: List[HttpMethod]
    opts: Optional[MethodOptions]

    def __init__(
        self,
        route: Route,
        methods: List[HttpMethod],
        *middleware: HttpMiddleware | HttpHandler,
        opts: MethodOptions,
    ):
        """Construct a method handler for the specified route."""
        self.route = route
        self.methods = methods

        handler = compose_middleware(*middleware)

        self.server = ApiRouteWorker(
            api_name=self.route.api.name, path=self.route.path, methods=self.methods, handler=handler, options=opts
        )


def _http_context_from_proto(msg: ProtoHttpRequest) -> HttpContext:
    """Construct a new HttpContext from a Http trigger from the Nitric Membrane."""
    headers: Record = {k: v.value for (k, v) in msg.headers.items()}
    query: Record = {k: v.value for (k, v) in msg.query_params.items()}

    return HttpContext(
        request=HttpRequest(
            data=msg.body,
            method=msg.method,
            query=query,
            path=msg.path,
            params=dict(msg.path_params.items()),
            headers=headers,
        )
    )


def _http_context_to_proto_response(ctx: HttpContext) -> ProtoHttpResponse:
    """Construct a HttpResponse for the Nitric Membrane from this context object."""
    body = ctx.res.body if ctx.res.body else bytes()
    headers: Dict[str, HeaderValue] = {}
    for k, v in ctx.res.headers.items():
        hv = HeaderValue()
        hv.value = HttpContext._ensure_value_is_list(v)  # pylint: disable=protected-access
        headers[k] = hv

    return ProtoHttpResponse(
        status=ctx.res.status,
        body=body,
        headers=headers,
    )


class ApiRouteWorker(FunctionServer):
    """A worker for handling HTTP requests for a specific API route."""

    _handler: HttpHandler
    _registration_request: RegistrationRequest
    _responses: AsyncNotifierList[ClientMessage]
    _options: MethodOptions

    def __init__(
        self,
        api_name: str,
        path: str,
        methods: List[HttpMethod],
        handler: HttpHandler,
        options: MethodOptions,
    ):
        """Construct a new ApiRouteWorker."""
        sec = {opt.name: ApiWorkerScopes(scopes=opt.scopes) for opt in options.security} if options.security else {}
        reg_options = ApiWorkerOptions(
            security=sec,
            security_disabled=True if options.security == [] else False,
        )

        self._handler = handler
        self._responses = AsyncNotifierList()
        self._options = options
        self._registration_request = RegistrationRequest(
            api=api_name,
            path=path,
            methods=[method.value for method in methods],
            options=reg_options,
        )

        Nitric._register_worker(self)

    async def _route_request_iterator(self):
        # Register with the server
        yield ClientMessage(registration_request=self._registration_request)
        # wait for any responses for the server and send them
        async for response in self._responses:
            yield response

    async def start(self) -> None:
        """Register this API route handler and handle http requests."""
        channel = new_default_channel()
        server = ApiStub(channel=channel)

        # Attach security rules for this route
        for security_rule in self._options.security if self._options.security else []:
            _attach_oidc(api_name=self._registration_request.api, options=security_rule)

        try:
            async for server_msg in server.serve(self._route_request_iterator()):
                msg_type, _ = betterproto.which_one_of(server_msg, "content")

                if msg_type == "registration_response":
                    continue
                if msg_type == "http_request":
                    ctx = _http_context_from_proto(server_msg.http_request)
                    response: ClientMessage
                    try:
                        result = await self._handler(ctx)
                        ctx = result if result else ctx
                        response = ClientMessage(id=server_msg.id, http_response=_http_context_to_proto_response(ctx))
                    except Exception as e:  # pylint: disable=broad-except
                        logging.exception("An unhandled error occurred in an api route handler: %s", e)
                        failed_http_response = ProtoHttpResponse(
                            status=500,
                            body=b"Internal Server Error",
                        )
                        response = ClientMessage(id=server_msg.id, http_response=failed_http_response)
                    await self._responses.add_item(response)
        except grpclib.exceptions.GRPCError as e:
            print(f"Stream terminated: {e.message}")
        except grpclib.exceptions.StreamTerminatedError:
            print("Stream from membrane closed.")
        finally:
            print("Closing client stream")
            channel.close()


def api(name: str, opts: Optional[ApiOptions] = None) -> Api:
    """Create a new API resource."""
    return Nitric._create_resource(Api, name, opts=opts)  # type: ignore pylint: disable=protected-access


class OidcOptions:
    """An OIDC security definition for an API."""

    name: str
    issuer: str
    audiences: List[str]

    def __init__(self, name: str, issuer: str, audiences: List[str]):
        """Construct a new OIDC security definition."""
        self.name = name
        self.issuer = issuer
        self.audiences = audiences


class ScopedOidcOptions(OidcOptions):
    """An OIDC security definition for an API with scopes."""

    scopes: List[str]

    def __init__(self, name: str, issuer: str, audiences: List[str], scopes: List[str]):
        """Construct a new scoped OIDC security definition."""
        super().__init__(name=name, issuer=issuer, audiences=audiences)
        self.scopes = scopes


def _oidc_to_resource(b: OidcSecurityDefinition) -> ResourceIdentifier:
    """Generate a resource identifier for an OIDC security definition."""
    return ResourceIdentifier(name=b.name, type=ResourceType.ApiSecurityDefinition)


class OidcSecurityDefinition(BaseResource):
    """An OIDC security definition for an API."""

    api_name: str
    issuer: str
    rule_name: str
    audiences: List[str]

    def __init__(self, name: str, api_name: str, options: OidcOptions):
        """Construct a new OIDC security definition."""
        super().__init__(name)
        self.api_name = api_name
        self.issuer = options.issuer
        self.audiences = options.audiences
        self.rule_name = options.name

    async def _register(self) -> None:
        try:
            await self._resources_stub.declare(
                resource_declare_request=ResourceDeclareRequest(
                    id=_oidc_to_resource(self),
                    api_security_definition=ApiSecurityDefinitionResource(
                        api_name=self.api_name,
                        oidc=ApiOpenIdConnectionDefinition(
                            issuer=self.issuer,
                            audiences=self.audiences,
                        ),
                    ),
                )
            )
        except GRPCError as grpc_err:
            raise exception_from_grpc_error(grpc_err) from grpc_err


def _attach_oidc(api_name: str, options: OidcOptions) -> OidcSecurityDefinition:
    return Nitric._create_resource(  # pylint: disable=protected-access
        OidcSecurityDefinition, f"{options.name}-{api_name}", api_name, options
    )


def oidc_rule(name: str, issuer: str, audiences: List[str]) -> Callable[[List[str]], ScopedOidcOptions]:
    """Create a function that returns scoped OIDC security definitions."""
    return lambda scopes: ScopedOidcOptions(name, issuer, audiences, scopes)



#
# Copyright (c) 2021 Nitric Technologies Pty Ltd.
#
# This file is part of Nitric Python 3 SDK.
# See https://github.com/nitrictech/python-sdk for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

import logging
from dataclasses import dataclass
from datetime import timedelta
from enum import Enum
from typing import Callable, List, Literal, Optional, Union, cast
from warnings import warn

import betterproto
import grpclib
from grpclib import GRPCError
from grpclib.client import Channel

from nitric.application import Nitric
from nitric.bidi import AsyncNotifierList
from nitric.context import FunctionServer, Handler, Middleware
from nitric.exception import InvalidArgumentException, exception_from_grpc_error
from nitric.proto.resources.v1 import Action, ResourceDeclareRequest, ResourceIdentifier, ResourceType
from nitric.proto.storage.v1 import (
    BlobEventRequest,
    BlobEventResponse,
    BlobEventType,
    ClientMessage,
    RegistrationRequest,
    StorageDeleteRequest,
    StorageExistsRequest,
    StorageListBlobsRequest,
    StorageListenerStub,
    StoragePreSignUrlRequest,
    StoragePreSignUrlRequestOperation,
    StorageReadRequest,
    StorageStub,
    StorageWriteRequest,
)
from nitric.resources.resource import SecureResource
from nitric.utils import new_default_channel


class BucketNotifyRequest:
    """Represents a translated Event, from a subscribed bucket notification, forwarded from the Nitric Membrane."""

    bucket_name: str
    key: str
    notification_type: BlobEventType
    bucket: BucketRef
    file: FileRef

    def __init__(self, bucket_name: str, key: str, notification_type: BlobEventType):
        """Construct a new BucketNotifyRequest."""
        self.bucket_name = bucket_name
        self.key = key
        self.notification_type = notification_type
        self.bucket = BucketRef(bucket_name)
        self.file = self.bucket.file(key)


class BucketNotifyResponse:
    """Represents the response to a trigger from a Bucket."""

    def __init__(self, success: bool = True):
        """Construct a new BucketNotificationResponse."""
        self.success = success


class BucketNotificationContext:
    """Represents the full request/response context for a bucket notification trigger."""

    def __init__(self, request: BucketNotifyRequest, response: Optional[BucketNotifyResponse] = None):
        """Construct a new BucketNotificationContext."""
        self.req = request
        self.res = response if response else BucketNotifyResponse()


class FileNotifyRequest(BucketNotifyRequest):
    """Represents a translated Event, from a subscribed bucket notification, forwarded from the Nitric Membrane."""

    def __init__(
        self,
        bucket_name: str,
        bucket_ref: BucketRef,
        key: str,
        notification_type: BlobEventType,
    ):
        """Construct a new FileNotificationRequest."""
        super().__init__(bucket_name=bucket_name, key=key, notification_type=notification_type)
        self.file = bucket_ref.file(key)


class FileNotificationContext(BucketNotificationContext):
    """Represents the full request/response context for a bucket notification trigger."""

    def __init__(self, request: FileNotifyRequest, response: Optional[BucketNotifyResponse] = None):
        """Construct a new FileNotificationContext."""
        super().__init__(request=request, response=response)
        self.req = request

    @staticmethod
    def _from_client_message_with_bucket(msg: BlobEventRequest, bucket_ref) -> FileNotificationContext:
        """Construct a new FileNotificationTrigger from a Bucket Notification trigger from the Nitric Membrane."""
        return FileNotificationContext(
            request=FileNotifyRequest(
                bucket_name=msg.bucket_name,
                key=msg.blob_event.key,
                bucket_ref=bucket_ref,
                notification_type=msg.blob_event.type,
            )
        )


BucketNotificationMiddleware = Middleware[BucketNotificationContext]
BucketNotificationHandler = Handler[BucketNotificationContext]

FileNotificationMiddleware = Middleware[FileNotificationContext]
FileNotificationHandler = Handler[FileNotificationContext]


class BucketRef(object):
    """A reference to a deployed storage bucket, used to interact with the bucket at runtime."""

    _channel: Channel
    _storage_stub: StorageStub
    name: str

    def __init__(self, name: str):
        """Construct a Nitric Storage Client."""
        self._channel: Union[Channel, None] = new_default_channel()
        self._storage_stub = StorageStub(channel=self._channel)
        self.name = name

    def __del__(self):
        # close the channel when this client is destroyed
        if self._channel is not None:
            self._channel.close()

    def file(self, key: str):
        """Return a reference to a file in this bucket."""
        return FileRef(_bucket=self, key=key)

    async def files(self):
        """Return a list of files in this bucket."""
        resp = await self._storage_stub.list_blobs(
            storage_list_blobs_request=StorageListBlobsRequest(bucket_name=self.name)
        )
        return [self.file(f.key) for f in resp.files]

    async def exists(self, key: str) -> bool:
        """Return true if a file in the bucket exists."""
        resp = await self._storage_stub.exists(
            storage_exists_request=StorageExistsRequest(bucket_name=self.name, key=key)
        )
        return resp.exists

    def on(
        self, notification_type: str, notification_prefix_filter: str
    ) -> Callable[[BucketNotificationHandler], None]:
        """Create and return a bucket notification decorator for this bucket."""

        def decorator(func: BucketNotificationHandler) -> None:
            Listener(
                bucket_name=self.name,
                notification_type=notification_type,
                notification_prefix_filter=notification_prefix_filter,
                handler=func,
            )

        return decorator


class FileMode(Enum):
    """Definition of available operation modes for file signed URLs."""

    READ = 0
    WRITE = 1

    def to_request_operation(self) -> StoragePreSignUrlRequestOperation:
        """Convert FileMode to a StoragePreSignUrlRequestOperation."""
        if self == FileMode.READ:
            return StoragePreSignUrlRequestOperation.READ
        elif self == FileMode.WRITE:
            return StoragePreSignUrlRequestOperation.WRITE
        else:
            raise InvalidArgumentException("Invalid FileMode")


@dataclass(frozen=True, order=True)
class FileRef(object):
    """A reference to a file in a bucket, used to perform operations on that file."""

    _bucket: BucketRef
    key: str

    async def write(self, body: bytes):
        """
        Write the bytes as the content of this file.

        Will create the file if it doesn't already exist.
        """
        try:
            await self._bucket._storage_stub.write(
                storage_write_request=StorageWriteRequest(bucket_name=self._bucket.name, key=self.key, body=body)
            )
        except GRPCError as grpc_err:
            raise exception_from_grpc_error(grpc_err) from grpc_err

    async def read(self) -> bytes:
        """Read this files contents from the bucket."""
        try:
            response = await self._bucket._storage_stub.read(
                storage_read_request=StorageReadRequest(bucket_name=self._bucket.name, key=self.key)
            )
            return response.body
        except GRPCError as grpc_err:
            raise exception_from_grpc_error(grpc_err) from grpc_err

    async def delete(self):
        """Delete this file from the bucket."""
        try:
            await self._bucket._storage_stub.delete(
                storage_delete_request=StorageDeleteRequest(bucket_name=self._bucket.name, key=self.key)
            )
        except GRPCError as grpc_err:
            raise exception_from_grpc_error(grpc_err) from grpc_err

    async def upload_url(self, expiry: Optional[Union[timedelta, int]] = None):
        """
        Get a temporary writable URL to this file.

        Parameters
        ----------
        expiry : int, timedelta, optional
            The expiry time for the signed URL.
            If an integer is provided, it is treated as seconds. Default is 600 seconds.

        Returns
        -------
        str: The signed URL.

        """
        return await self._sign_url(mode=FileMode.WRITE, expiry=expiry)

    async def download_url(self, expiry: Optional[Union[timedelta, int]] = None):
        """
        Get a temporary readable URL to this file.

        Parameters
        ----------
        expiry : int, timedelta, optional
            The expiry time for the signed URL.
            If an integer is provided, it is treated as seconds. Default is 600 seconds.

        Returns
        -------
        str: The signed URL.

        """
        return await self._sign_url(mode=FileMode.READ, expiry=expiry)

    async def _sign_url(self, mode: FileMode = FileMode.READ, expiry: Optional[Union[timedelta, int]] = None):
        """Generate a signed URL for reading or writing to a file."""
        if expiry is None:
            expiry = timedelta(seconds=600)
        if not isinstance(expiry, timedelta):
            expiry = timedelta(seconds=expiry)

        try:
            response = await self._bucket._storage_stub.pre_sign_url(
                storage_pre_sign_url_request=StoragePreSignUrlRequest(
                    bucket_name=self._bucket.name, key=self.key, operation=mode.to_request_operation(), expiry=expiry
                )
            )
            return response.url
        except GRPCError as grpc_err:
            raise exception_from_grpc_error(grpc_err) from grpc_err


LegacyBucketPermission = Literal["reading", "writing", "deleting"]
BucketPermission = Literal["read", "write", "delete"]

legacy_perms: List[LegacyBucketPermission] = ["reading", "writing", "deleting"]
new_perms: List[BucketPermission] = ["read", "write", "delete"]


def check_permission(permission: Union[LegacyBucketPermission, BucketPermission]) -> BucketPermission:
    """Check if the permission is valid and return the new permission if it is a legacy permission."""
    if permission in legacy_perms:
        new_perm = new_perms[legacy_perms.index(cast(LegacyBucketPermission, permission))]
        warn(
            f"The permission '{permission}' is deprecated. Use '{new_perm}' instead.", DeprecationWarning, stacklevel=2
        )
        return new_perm
    elif permission in new_perms:
        return cast(BucketPermission, permission)
    else:
        raise ValueError("Invalid permission value, must be one of 'read', 'write', or 'delete'.")


class BucketNotificationWorkerOptions:
    """Options for bucket notification workers."""

    def __init__(self, bucket_name: str, notification_type: str, notification_prefix_filter: str):
        """Construct a new options object."""
        self.bucket_name = bucket_name
        self.notification_type = BucketNotificationWorkerOptions._to_grpc_event_type(notification_type.lower())
        self.notification_prefix_filter = notification_prefix_filter

    @staticmethod
    def _to_grpc_event_type(event_type: str) -> BlobEventType:
        if event_type == "write":
            return BlobEventType.Created
        elif event_type == "delete":
            return BlobEventType.Deleted
        else:
            raise ValueError(f"Event type {event_type} is unsupported")


class Bucket(SecureResource):
    """A bucket resource, used for storage and retrieval of blob/binary data."""

    name: str
    actions: List[Action]
    _server: FunctionServer

    async def _register(self) -> None:
        try:
            await self._resources_stub.declare(
                resource_declare_request=ResourceDeclareRequest(id=self._to_resource_id())
            )
        except GRPCError as grpc_err:
            raise exception_from_grpc_error(grpc_err) from grpc_err

    def _perms_to_actions(self, *args: BucketPermission) -> List[Action]:
        permission_actions_map: dict[BucketPermission, List[Action]] = {
            "read": [Action.BucketFileGet, Action.BucketFileList],
            "write": [Action.BucketFilePut],
            "delete": [Action.BucketFileDelete],
        }

        return [action for perm in args for action in permission_actions_map[perm]]

    def _to_resource_id(self) -> ResourceIdentifier:
        return ResourceIdentifier(name=self.name, type=ResourceType.Bucket)

    def allow(
        self,
        perm: Union[LegacyBucketPermission, BucketPermission],
        *args: Union[LegacyBucketPermission, BucketPermission],
    ) -> BucketRef:
        """Request the required permissions for this resource."""
        all_perms: List[BucketPermission] = [check_permission(perm)] + [check_permission(p) for p in args]

        str_args = [str(permission) for permission in all_perms]
        self._register_policy(*str_args)

        return BucketRef(self.name)

    def on(
        self, notification_type: str, notification_prefix_filter: str
    ) -> Callable[[BucketNotificationHandler], None]:
        """Create and return a bucket notification decorator for this bucket."""

        def decorator(func: BucketNotificationHandler) -> None:
            Listener(
                bucket_name=self.name,
                notification_type=notification_type,
                notification_prefix_filter=notification_prefix_filter,
                handler=func,
            )

        return decorator


class Listener(FunctionServer):
    """A bucket event listener."""

    _handler: BucketNotificationHandler
    _registration_request: RegistrationRequest
    _responses: AsyncNotifierList[ClientMessage]

    def __init__(
        self,
        bucket_name: str,
        notification_type: str,
        notification_prefix_filter: str,
        handler: BucketNotificationHandler,
    ):
        """Construct a new bucket event listener."""
        self._handler = handler
        self._responses = AsyncNotifierList()

        event_type = BlobEventType.Created
        if "del" in notification_type:
            event_type = BlobEventType.Deleted

        self._registration_request = RegistrationRequest(
            bucket_name=bucket_name,
            blob_event_type=event_type,
            key_prefix_filter=notification_prefix_filter,
        )

        # noinspection PyProtectedMember
        Nitric._register_worker(self)

    async def _listener_request_iterator(self):
        # Register with the server
        yield ClientMessage(registration_request=self._registration_request)
        # wait for any responses for the server and send them
        async for response in self._responses:
            yield response

    async def start(self) -> None:
        """Register this bucket listener and listen for events."""
        channel = new_default_channel()
        server = StorageListenerStub(channel=channel)

        try:
            async for server_msg in server.listen(self._listener_request_iterator()):
                msg_type, _ = betterproto.which_one_of(server_msg, "content")

                if msg_type == "registration_response":
                    continue
                if msg_type == "blob_event_request":
                    ctx = BucketNotificationContext(
                        request=BucketNotifyRequest(
                            bucket_name=server_msg.blob_event_request.bucket_name,
                            key=server_msg.blob_event_request.blob_event.key,
                            notification_type=server_msg.blob_event_request.blob_event.type,
                        )
                    )
                    response: ClientMessage
                    try:
                        result = await self._handler(ctx)
                        ctx = result if result else ctx
                        be = BlobEventResponse(success=ctx.res.success)
                        response = ClientMessage(id=server_msg.id, blob_event_response=be)
                    except Exception as e:  # pylint: disable=broad-except
                        logging.exception("An unhandled error occurred in a bucket event listener: %s", e)
                        be = BlobEventResponse(success=False)
                        response = ClientMessage(id=server_msg.id, blob_event_response=be)
                    await self._responses.add_item(response)
        except grpclib.exceptions.GRPCError as e:
            print(f"Stream terminated: {e.message}")
        except grpclib.exceptions.StreamTerminatedError:
            print("Stream from membrane closed.")
        except KeyboardInterrupt:
            print("Keyboard interrupt")
        finally:
            print("Closing client stream")
            channel.close()
        print("Listener stopped")


def bucket(name: str) -> Bucket:
    """
    Create and register a bucket.

    If a bucket has already been registered with the same name, the original reference will be reused.
    """
    return Nitric._create_resource(Bucket, name)



#
# Copyright (c) 2021 Nitric Technologies Pty Ltd.
#
# This file is part of Nitric Python 3 SDK.
# See https://github.com/nitrictech/python-sdk for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: nitric/proto/deployments/v1/deployments.proto
# plugin: python-betterproto
# This file has been @generated
import warnings
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    AsyncIterator,
    Dict,
    List,
    Optional,
)

import betterproto
import betterproto.lib.google.protobuf as betterproto_lib_google_protobuf
import grpclib
from betterproto.grpc.grpclib_server import ServiceBase

from ...resources import v1 as __resources_v1__
from ...storage import v1 as __storage_v1__


if TYPE_CHECKING:
    import grpclib.server
    from betterproto.grpc.grpclib_client import MetadataLike
    from grpclib.metadata import Deadline


class ResourceDeploymentAction(betterproto.Enum):
    CREATE = 0
    """A new resource is being created"""

    UPDATE = 1
    """An existing resource is being updated"""

    REPLACE = 2
    """An existing resource is being replaced"""

    SAME = 3
    """No-op on the resource (it already exists and requires no changes)"""

    DELETE = 4
    """An existing resource is being deleted"""


class ResourceDeploymentStatus(betterproto.Enum):
    PENDING = 0
    """The action hasn't started, usually due to a dependency"""

    IN_PROGRESS = 1
    """
    The action in currently in-flight, e.g. waiting for cloud provder to
    provision a resource
    """

    SUCCESS = 2
    """The action has been applied successfully"""

    FAILED = 3
    """The action has failed to be (completely) applied"""


@dataclass(eq=False, repr=False)
class DeploymentUpRequest(betterproto.Message):
    spec: "Spec" = betterproto.message_field(1)
    """The spec to deploy"""

    attributes: "betterproto_lib_google_protobuf.Struct" = betterproto.message_field(2)
    """
    A map of attributes related to the deploy request this allows for adding
    project identifiers etc.
    """

    interactive: bool = betterproto.bool_field(3)
    """
    A hint to the provider of the kind of output that the client can accept
    This will allow provider developers to provider richer output back to
    clients.
    """


@dataclass(eq=False, repr=False)
class DeploymentUpEvent(betterproto.Message):
    message: str = betterproto.string_field(1, group="content")
    update: "ResourceUpdate" = betterproto.message_field(2, group="content")
    result: "UpResult" = betterproto.message_field(3, group="content")


@dataclass(eq=False, repr=False)
class ResourceUpdate(betterproto.Message):
    id: "__resources_v1__.ResourceIdentifier" = betterproto.message_field(1)
    """
    The resource being updated, if this is nil the update applies to the stack
    """

    action: "ResourceDeploymentAction" = betterproto.enum_field(3)
    """The type of update being applied"""

    status: "ResourceDeploymentStatus" = betterproto.enum_field(4)
    """The current status of the action being applied"""

    sub_resource: str = betterproto.string_field(5)
    """
    (optional) A globally unique identifier (scoped to the id above), used when
    Nitric Resources map 1:many in a cloud provider. e.g. the container image
    repository for a service deployment. This can also be set when id is nil
    above and it will imply a non-nitric resource that is necessary to deploy
    for a stack to operate  e.g. an Azure StorageAccount
    """

    message: str = betterproto.string_field(6)
    """Additional information about the update"""


@dataclass(eq=False, repr=False)
class UpResult(betterproto.Message):
    """Terminal message indicating deployment success"""

    success: bool = betterproto.bool_field(1)
    """Indicate the success status"""

    text: str = betterproto.string_field(2, group="content")
    """Simple text output as result"""


@dataclass(eq=False, repr=False)
class DeploymentDownRequest(betterproto.Message):
    attributes: "betterproto_lib_google_protobuf.Struct" = betterproto.message_field(1)
    """
    A map of attributes related to the deploy request this allows for adding
    project identifiers etc.
    """

    interactive: bool = betterproto.bool_field(2)
    """
    A hint to the provider of the kind of output that the client can accept
    This will allow provider developers to provider richer output back to
    clients.
    """


@dataclass(eq=False, repr=False)
class DeploymentDownEvent(betterproto.Message):
    message: str = betterproto.string_field(1, group="content")
    result: "DownResult" = betterproto.message_field(2, group="content")
    update: "ResourceUpdate" = betterproto.message_field(3, group="content")


@dataclass(eq=False, repr=False)
class DownResult(betterproto.Message):
    """Terminal message indicating deployment success"""

    pass


@dataclass(eq=False, repr=False)
class ImageSource(betterproto.Message):
    """An image source to be used for service deployment"""

    uri: str = betterproto.string_field(1)
    """
    URI of the docker image To support remote images this may also need to
    provide auth information
    """


@dataclass(eq=False, repr=False)
class Service(betterproto.Message):
    """A unit of compute (i.e. function/container)"""

    image: "ImageSource" = betterproto.message_field(1, group="source")
    """Container image as a service"""

    workers: int = betterproto.int32_field(10)
    """Expected worker count for this service"""

    timeout: int = betterproto.int32_field(11)
    """Configurable timeout for request handling"""

    memory: int = betterproto.int32_field(12)
    """Configurable memory size for this instance"""

    type: str = betterproto.string_field(13)
    """
    A simple type property describes the requested type of service that this
    should be for this project, a provider can implement how this request is
    satisfied in any way
    """

    env: Dict[str, str] = betterproto.map_field(
        14, betterproto.TYPE_STRING, betterproto.TYPE_STRING
    )
    """Environment variables for this service"""

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.is_set("timeout"):
            warnings.warn("Service.timeout is deprecated", DeprecationWarning)
        if self.is_set("memory"):
            warnings.warn("Service.memory is deprecated", DeprecationWarning)


@dataclass(eq=False, repr=False)
class Bucket(betterproto.Message):
    listeners: List["BucketListener"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class BucketListener(betterproto.Message):
    config: "__storage_v1__.RegistrationRequest" = betterproto.message_field(1)
    service: str = betterproto.string_field(2, group="target")
    """The name of an service to target"""


@dataclass(eq=False, repr=False)
class Topic(betterproto.Message):
    subscriptions: List["SubscriptionTarget"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class Queue(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class KeyValueStore(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class Secret(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class SubscriptionTarget(betterproto.Message):
    service: str = betterproto.string_field(1, group="target")
    """The name of an service to target"""


@dataclass(eq=False, repr=False)
class TopicSubscription(betterproto.Message):
    target: "SubscriptionTarget" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class HttpTarget(betterproto.Message):
    service: str = betterproto.string_field(1, group="target")
    """The name of an service to target"""


@dataclass(eq=False, repr=False)
class Http(betterproto.Message):
    """An http proxy resource"""

    target: "HttpTarget" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class Api(betterproto.Message):
    openapi: str = betterproto.string_field(1, group="document")
    """
    An OpenAPI document for deployment This document will contain extensions
    that hint of services that should be targeted as part of the deployment
    """


@dataclass(eq=False, repr=False)
class Websocket(betterproto.Message):
    """Declare a new websocket"""

    connect_target: "WebsocketTarget" = betterproto.message_field(1)
    """Target for handling new client connections"""

    disconnect_target: "WebsocketTarget" = betterproto.message_field(2)
    """Target for handling client disconnections"""

    message_target: "WebsocketTarget" = betterproto.message_field(3)
    """Target for handling all other message types"""


@dataclass(eq=False, repr=False)
class WebsocketTarget(betterproto.Message):
    service: str = betterproto.string_field(1, group="target")
    """The name of an service to target"""


@dataclass(eq=False, repr=False)
class ScheduleTarget(betterproto.Message):
    service: str = betterproto.string_field(1, group="target")
    """The name of an service to target"""


@dataclass(eq=False, repr=False)
class Schedule(betterproto.Message):
    target: "ScheduleTarget" = betterproto.message_field(1)
    every: "ScheduleEvery" = betterproto.message_field(10, group="cadence")
    cron: "ScheduleCron" = betterproto.message_field(11, group="cadence")


@dataclass(eq=False, repr=False)
class ScheduleEvery(betterproto.Message):
    rate: str = betterproto.string_field(1)
    """
    rate string e.g. '5 minutes'. Value frequencies are 'minutes', 'hours',
    'days'.
    """


@dataclass(eq=False, repr=False)
class ScheduleCron(betterproto.Message):
    expression: str = betterproto.string_field(1)
    """standard unix cron expression"""


@dataclass(eq=False, repr=False)
class Resource(betterproto.Message):
    id: "__resources_v1__.ResourceIdentifier" = betterproto.message_field(1)
    service: "Service" = betterproto.message_field(10, group="config")
    bucket: "Bucket" = betterproto.message_field(11, group="config")
    topic: "Topic" = betterproto.message_field(12, group="config")
    api: "Api" = betterproto.message_field(13, group="config")
    policy: "Policy" = betterproto.message_field(14, group="config")
    schedule: "Schedule" = betterproto.message_field(15, group="config")
    key_value_store: "KeyValueStore" = betterproto.message_field(16, group="config")
    secret: "Secret" = betterproto.message_field(17, group="config")
    websocket: "Websocket" = betterproto.message_field(18, group="config")
    http: "Http" = betterproto.message_field(19, group="config")
    queue: "Queue" = betterproto.message_field(20, group="config")


@dataclass(eq=False, repr=False)
class Policy(betterproto.Message):
    """
    This is already defined in the resource contracts, unfortunately there are
    parts we don't want to duplicate, such as API config
    """

    principals: List["Resource"] = betterproto.message_field(1)
    actions: List["__resources_v1__.Action"] = betterproto.enum_field(2)
    resources: List["Resource"] = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class Spec(betterproto.Message):
    resources: List["Resource"] = betterproto.message_field(1)
    """list of resources to deploy"""


class DeploymentStub(betterproto.ServiceStub):
    async def up(
        self,
        deployment_up_request: "DeploymentUpRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> AsyncIterator["DeploymentUpEvent"]:
        async for response in self._unary_stream(
            "/nitric.proto.deployments.v1.Deployment/Up",
            deployment_up_request,
            DeploymentUpEvent,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        ):
            yield response

    async def down(
        self,
        deployment_down_request: "DeploymentDownRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> AsyncIterator["DeploymentDownEvent"]:
        async for response in self._unary_stream(
            "/nitric.proto.deployments.v1.Deployment/Down",
            deployment_down_request,
            DeploymentDownEvent,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        ):
            yield response


class DeploymentBase(ServiceBase):
    async def up(
        self, deployment_up_request: "DeploymentUpRequest"
    ) -> AsyncIterator["DeploymentUpEvent"]:
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)
        yield DeploymentUpEvent()

    async def down(
        self, deployment_down_request: "DeploymentDownRequest"
    ) -> AsyncIterator["DeploymentDownEvent"]:
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)
        yield DeploymentDownEvent()

    async def __rpc_up(
        self, stream: "grpclib.server.Stream[DeploymentUpRequest, DeploymentUpEvent]"
    ) -> None:
        request = await stream.recv_message()
        await self._call_rpc_handler_server_stream(
            self.up,
            stream,
            request,
        )

    async def __rpc_down(
        self,
        stream: "grpclib.server.Stream[DeploymentDownRequest, DeploymentDownEvent]",
    ) -> None:
        request = await stream.recv_message()
        await self._call_rpc_handler_server_stream(
            self.down,
            stream,
            request,
        )

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/nitric.proto.deployments.v1.Deployment/Up": grpclib.const.Handler(
                self.__rpc_up,
                grpclib.const.Cardinality.UNARY_STREAM,
                DeploymentUpRequest,
                DeploymentUpEvent,
            ),
            "/nitric.proto.deployments.v1.Deployment/Down": grpclib.const.Handler(
                self.__rpc_down,
                grpclib.const.Cardinality.UNARY_STREAM,
                DeploymentDownRequest,
                DeploymentDownEvent,
            ),
        }



#
# Copyright (c) 2021 Nitric Technologies Pty Ltd.
#
# This file is part of Nitric Python 3 SDK.
# See https://github.com/nitrictech/python-sdk for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

import logging
from typing import Callable, Literal

import betterproto
import grpclib
from grpclib import GRPCError
from grpclib.client import Channel

from nitric.application import Nitric
from nitric.bidi import AsyncNotifierList
from nitric.context import (
    FunctionServer,
    Record,
    WebsocketConnectionRequest,
    WebsocketConnectionResponse,
    WebsocketContext,
    WebsocketHandler,
    WebsocketMessageRequest,
    WebsocketRequest,
)
from nitric.exception import exception_from_grpc_error
from nitric.proto.resources.v1 import Action, PolicyResource, ResourceDeclareRequest, ResourceIdentifier, ResourceType
from nitric.proto.websockets.v1 import ClientMessage, RegistrationRequest
from nitric.proto.websockets.v1 import WebsocketConnectionResponse as ProtoWebsocketConnectionResponse
from nitric.proto.websockets.v1 import (
    WebsocketEventRequest,
    WebsocketEventResponse,
    WebsocketEventType,
    WebsocketHandlerStub,
    WebsocketSendRequest,
    WebsocketStub,
)
from nitric.resources.resource import Resource as BaseResource
from nitric.utils import new_default_channel


class WebsocketRef:
    """A reference to a deployed websocket, used to interact with the websocket at runtime."""

    def __init__(self) -> None:
        """Construct a Nitric Websocket Client."""
        self._channel: Channel = new_default_channel()
        self._websocket_stub = WebsocketStub(channel=self._channel)

    async def send(self, socket: str, connection_id: str, data: bytes):
        """Send data to a connection on a socket."""
        try:
            await self._websocket_stub.send_message(
                websocket_send_request=WebsocketSendRequest(socket_name=socket, connection_id=connection_id, data=data)
            )
        except GRPCError as grpc_err:
            raise exception_from_grpc_error(grpc_err) from grpc_err


def _to_grpc_event_type(event_type: Literal["connect", "disconnect", "message"]) -> WebsocketEventType:
    if event_type == "connect":
        return WebsocketEventType.Connect
    elif event_type == "disconnect":
        return WebsocketEventType.Disconnect
    elif event_type == "message":
        return WebsocketEventType.Message
    else:
        raise ValueError(f"Event type {event_type} is unsupported")


def _to_resource(b: Websocket) -> ResourceIdentifier:
    return ResourceIdentifier(name=b.name, type=ResourceType.Websocket)


class Websocket(BaseResource):
    """A Websocket API."""

    app: Nitric
    _websocket: WebsocketRef
    name: str

    def __init__(self, name: str):
        """Construct a new Websocket API."""
        super().__init__(name)
        self._websocket = WebsocketRef()

    async def _register(self) -> None:
        try:
            resource = _to_resource(self)
            default_principle = ResourceIdentifier(type=ResourceType.Service)

            await self._resources_stub.declare(
                resource_declare_request=ResourceDeclareRequest(
                    id=resource,
                )
            )

            await self._resources_stub.declare(
                resource_declare_request=ResourceDeclareRequest(
                    id=ResourceIdentifier(type=ResourceType.Policy),
                    policy=PolicyResource(
                        actions=[Action.WebsocketManage], principals=[default_principle], resources=[resource]
                    ),
                )
            )

        except GRPCError as grpc_err:
            raise exception_from_grpc_error(grpc_err) from grpc_err

    async def send(
        self,
        connection_id: str,
        data: bytes,
    ) -> None:
        """Send a message to a connection on this socket."""
        await self._websocket.send(socket=self.name, connection_id=connection_id, data=data)

    def on(self, event_type: WebsocketEventType) -> Callable[[WebsocketHandler], None]:
        """Create and return a worker decorator for this socket."""

        def decorator(func: WebsocketHandler) -> None:
            WebsocketWorker(
                socket_name=self.name,
                event_type=event_type,
                handler=func,
            )

        return decorator


def websocket(name: str) -> Websocket:
    """
    Create and registers a websocket.

    If a websocket has already been registered with the same name, the original reference will be reused.
    """
    return Nitric._create_resource(Websocket, name)  # type: ignore pylint: disable=protected-access


def _websocket_context_from_proto(msg: WebsocketEventRequest) -> WebsocketContext:
    """Construct a new WebsocketContext from a websocket trigger from the Nitric Server."""
    query: Record = {k: v.value for (k, v) in msg.connection.query_params.items()}

    req = WebsocketRequest(
        connection_id=msg.connection_id,
    )
    evt_type, _ = betterproto.which_one_of(msg, "websocket_event")
    if evt_type == "connection":
        req = WebsocketConnectionRequest(
            connection_id=msg.connection_id,
            query=query,
        )
    if evt_type == "message":
        req = WebsocketMessageRequest(connection_id=msg.connection_id, data=msg.message.body)

    return WebsocketContext(request=req)


def _websocket_context_to_proto_response(ctx: WebsocketContext) -> WebsocketEventResponse:
    resp = WebsocketEventResponse()
    if isinstance(ctx.res, WebsocketConnectionResponse):
        resp.connection_response = ProtoWebsocketConnectionResponse(reject=ctx.res.reject)
    return resp


class WebsocketWorker(FunctionServer):
    """A handler for websocket events."""

    _handler: WebsocketHandler
    _registration_request: RegistrationRequest
    _responses: AsyncNotifierList[ClientMessage]

    def __init__(
        self, socket_name: str, event_type: Literal["connect", "disconnect", "message"], handler: WebsocketHandler
    ):
        """Construct a new WebsocketHandler."""
        self._handler = handler
        self._responses = AsyncNotifierList()
        self._registration_request = RegistrationRequest(
            socket_name=socket_name, event_type=_to_grpc_event_type(event_type)
        )

        Nitric._register_worker(self)

    async def _ws_request_iterator(self):
        # Register with the server
        yield ClientMessage(registration_request=self._registration_request)
        # wait for any responses for the server and send them
        async for response in self._responses:
            yield response

    async def start(self) -> None:
        """Register this websocket handler and listen for messages."""
        channel = new_default_channel()
        server = WebsocketHandlerStub(channel=channel)

        try:
            async for server_msg in server.handle_events(self._ws_request_iterator()):
                msg_type, _ = betterproto.which_one_of(server_msg, "content")

                if msg_type == "registration_response":
                    continue
                if msg_type == "websocket_event_request":
                    ctx = _websocket_context_from_proto(server_msg.websocket_event_request)

                    response: ClientMessage
                    try:
                        result = await self._handler(ctx)
                        ctx = result if result else ctx
                        response = ClientMessage(id=server_msg.id, websocket_event_response=WebsocketEventResponse())
                        if isinstance(ctx.res, WebsocketConnectionResponse):
                            response.websocket_event_response.connection_response.reject = ctx.res.reject
                    except Exception as e:  # pylint: disable=broad-except
                        logging.exception("An unhandled error occurred in a websocket event handler: %s", e)
                        response = ClientMessage(id=server_msg.id, websocket_event_response=WebsocketEventResponse())
                        if isinstance(ctx.req, WebsocketConnectionRequest):
                            response.websocket_event_response.connection_response.reject = True
                    await self._responses.add_item(response)
        except grpclib.exceptions.GRPCError as e:
            print(f"Stream terminated: {e.message}")
        except grpclib.exceptions.StreamTerminatedError:
            print("Stream from membrane closed.")
        finally:
            print("Closing client stream")
            channel.close()



#
# Copyright (c) 2021 Nitric Technologies Pty Ltd.
#
# This file is part of Nitric Python 3 SDK.
# See https://github.com/nitrictech/python-sdk for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: nitric/proto/schedules/v1/schedules.proto
# plugin: python-betterproto
# This file has been @generated

from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    AsyncIterable,
    AsyncIterator,
    Dict,
    Iterable,
    Optional,
    Union,
)

import betterproto
import grpclib
from betterproto.grpc.grpclib_server import ServiceBase


if TYPE_CHECKING:
    import grpclib.server
    from betterproto.grpc.grpclib_client import MetadataLike
    from grpclib.metadata import Deadline


@dataclass(eq=False, repr=False)
class ClientMessage(betterproto.Message):
    """ClientMessages are sent from the service to the nitric server"""

    id: str = betterproto.string_field(1)
    """globally unique ID of the request/response pair"""

    registration_request: "RegistrationRequest" = betterproto.message_field(
        2, group="content"
    )
    """Register new a schedule"""

    interval_response: "IntervalResponse" = betterproto.message_field(
        3, group="content"
    )
    """
    Response to a schedule interval (i.e. response from callback function)
    """


@dataclass(eq=False, repr=False)
class IntervalRequest(betterproto.Message):
    schedule_name: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class ServerMessage(betterproto.Message):
    """ServerMessages are sent from the nitric server to the service"""

    id: str = betterproto.string_field(1)
    """globally unique ID of the request/response pair"""

    registration_response: "RegistrationResponse" = betterproto.message_field(
        2, group="content"
    )
    """Response to a schedule subscription request"""

    interval_request: "IntervalRequest" = betterproto.message_field(3, group="content")
    """A schedule interval trigger request (i.e. call the callback)"""


@dataclass(eq=False, repr=False)
class RegistrationRequest(betterproto.Message):
    schedule_name: str = betterproto.string_field(1)
    every: "ScheduleEvery" = betterproto.message_field(10, group="cadence")
    cron: "ScheduleCron" = betterproto.message_field(11, group="cadence")


@dataclass(eq=False, repr=False)
class ScheduleEvery(betterproto.Message):
    rate: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class ScheduleCron(betterproto.Message):
    expression: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class RegistrationResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class IntervalResponse(betterproto.Message):
    pass


class SchedulesStub(betterproto.ServiceStub):
    async def schedule(
        self,
        client_message_iterator: Union[
            AsyncIterable["ClientMessage"], Iterable["ClientMessage"]
        ],
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> AsyncIterator["ServerMessage"]:
        async for response in self._stream_stream(
            "/nitric.proto.schedules.v1.Schedules/Schedule",
            client_message_iterator,
            ClientMessage,
            ServerMessage,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        ):
            yield response


class SchedulesBase(ServiceBase):
    async def schedule(
        self, client_message_iterator: AsyncIterator["ClientMessage"]
    ) -> AsyncIterator["ServerMessage"]:
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)
        yield ServerMessage()

    async def __rpc_schedule(
        self, stream: "grpclib.server.Stream[ClientMessage, ServerMessage]"
    ) -> None:
        request = stream.__aiter__()
        await self._call_rpc_handler_server_stream(
            self.schedule,
            stream,
            request,
        )

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/nitric.proto.schedules.v1.Schedules/Schedule": grpclib.const.Handler(
                self.__rpc_schedule,
                grpclib.const.Cardinality.STREAM_STREAM,
                ClientMessage,
                ServerMessage,
            ),
        }



#
# Copyright (c) 2021 Nitric Technologies Pty Ltd.
#
# This file is part of Nitric Python 3 SDK.
# See https://github.com/nitrictech/python-sdk for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: nitric/proto/apis/v1/apis.proto
# plugin: python-betterproto
# This file has been @generated

from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    AsyncIterable,
    AsyncIterator,
    Dict,
    Iterable,
    List,
    Optional,
    Union,
)

import betterproto
import grpclib
from betterproto.grpc.grpclib_server import ServiceBase


if TYPE_CHECKING:
    import grpclib.server
    from betterproto.grpc.grpclib_client import MetadataLike
    from grpclib.metadata import Deadline


@dataclass(eq=False, repr=False)
class ApiDetailsRequest(betterproto.Message):
    api_name: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class ApiDetailsResponse(betterproto.Message):
    url: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class ClientMessage(betterproto.Message):
    """ClientMessage sent by the service to the nitric server"""

    id: str = betterproto.string_field(1)
    """globally unique ID of the request/response pair"""

    registration_request: "RegistrationRequest" = betterproto.message_field(
        2, group="content"
    )
    """Register an API route handler"""

    http_response: "HttpResponse" = betterproto.message_field(3, group="content")
    """Response to an HTTP request"""


@dataclass(eq=False, repr=False)
class HeaderValue(betterproto.Message):
    value: List[str] = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class QueryValue(betterproto.Message):
    value: List[str] = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class HttpRequest(betterproto.Message):
    method: str = betterproto.string_field(1)
    """The request method"""

    path: str = betterproto.string_field(2)
    """The path of the request"""

    headers: Dict[str, "HeaderValue"] = betterproto.map_field(
        3, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )
    """HTTP request headers"""

    query_params: Dict[str, "QueryValue"] = betterproto.map_field(
        4, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )
    """HTTP Query params"""

    path_params: Dict[str, str] = betterproto.map_field(
        5, betterproto.TYPE_STRING, betterproto.TYPE_STRING
    )
    """HTTP Path parameters"""

    body: bytes = betterproto.bytes_field(6)
    """HTTP Request body"""


@dataclass(eq=False, repr=False)
class HttpResponse(betterproto.Message):
    """HttpResponseMessage"""

    status: int = betterproto.int32_field(1)
    """The HTTP response status code"""

    headers: Dict[str, "HeaderValue"] = betterproto.map_field(
        2, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )
    """HTTP response headers"""

    body: bytes = betterproto.bytes_field(3)
    """HTTP response body"""


@dataclass(eq=False, repr=False)
class ServerMessage(betterproto.Message):
    """ServerMessage sent by the nitric server to the service"""

    id: str = betterproto.string_field(1)
    """globally unique ID of the request/response pair"""

    registration_response: "RegistrationResponse" = betterproto.message_field(
        2, group="content"
    )
    """Response to an API serve request"""

    http_request: "HttpRequest" = betterproto.message_field(3, group="content")
    """HTTP request to be routed to the service (handler)"""


@dataclass(eq=False, repr=False)
class RegistrationResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class ApiWorkerScopes(betterproto.Message):
    scopes: List[str] = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class ApiWorkerOptions(betterproto.Message):
    security: Dict[str, "ApiWorkerScopes"] = betterproto.map_field(
        1, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )
    """
    Apply security definitions to this operation This will be mapped to named
    ApiSecurityDefinitionResource(s)
    """

    security_disabled: bool = betterproto.bool_field(2)
    """
    explicitly disable security for this endpoint We need to do this as the
    default value of a repeated field is always empty so there is no way of
    knowing if security is explicitly disabled
    """


@dataclass(eq=False, repr=False)
class RegistrationRequest(betterproto.Message):
    api: str = betterproto.string_field(1)
    path: str = betterproto.string_field(2)
    methods: List[str] = betterproto.string_field(3)
    options: "ApiWorkerOptions" = betterproto.message_field(4)


class ApiStub(betterproto.ServiceStub):
    async def serve(
        self,
        client_message_iterator: Union[
            AsyncIterable["ClientMessage"], Iterable["ClientMessage"]
        ],
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> AsyncIterator["ServerMessage"]:
        async for response in self._stream_stream(
            "/nitric.proto.apis.v1.Api/Serve",
            client_message_iterator,
            ClientMessage,
            ServerMessage,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        ):
            yield response

    async def api_details(
        self,
        api_details_request: "ApiDetailsRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "ApiDetailsResponse":
        return await self._unary_unary(
            "/nitric.proto.apis.v1.Api/ApiDetails",
            api_details_request,
            ApiDetailsResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )


class ApiBase(ServiceBase):
    async def serve(
        self, client_message_iterator: AsyncIterator["ClientMessage"]
    ) -> AsyncIterator["ServerMessage"]:
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)
        yield ServerMessage()

    async def api_details(
        self, api_details_request: "ApiDetailsRequest"
    ) -> "ApiDetailsResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_serve(
        self, stream: "grpclib.server.Stream[ClientMessage, ServerMessage]"
    ) -> None:
        request = stream.__aiter__()
        await self._call_rpc_handler_server_stream(
            self.serve,
            stream,
            request,
        )

    async def __rpc_api_details(
        self, stream: "grpclib.server.Stream[ApiDetailsRequest, ApiDetailsResponse]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.api_details(request)
        await stream.send_message(response)

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/nitric.proto.apis.v1.Api/Serve": grpclib.const.Handler(
                self.__rpc_serve,
                grpclib.const.Cardinality.STREAM_STREAM,
                ClientMessage,
                ServerMessage,
            ),
            "/nitric.proto.apis.v1.Api/ApiDetails": grpclib.const.Handler(
                self.__rpc_api_details,
                grpclib.const.Cardinality.UNARY_UNARY,
                ApiDetailsRequest,
                ApiDetailsResponse,
            ),
        }



#
# Copyright (c) 2021 Nitric Technologies Pty Ltd.
#
# This file is part of Nitric Python 3 SDK.
# See https://github.com/nitrictech/python-sdk for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: nitric/proto/kvstore/v1/kvstore.proto
# plugin: python-betterproto
# This file has been @generated

from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    AsyncIterator,
    Dict,
    Optional,
)

import betterproto
import betterproto.lib.google.protobuf as betterproto_lib_google_protobuf
import grpclib
from betterproto.grpc.grpclib_server import ServiceBase


if TYPE_CHECKING:
    import grpclib.server
    from betterproto.grpc.grpclib_client import MetadataLike
    from grpclib.metadata import Deadline


@dataclass(eq=False, repr=False)
class Store(betterproto.Message):
    """Provides a Key/Value Store"""

    name: str = betterproto.string_field(1)
    """The store name"""


@dataclass(eq=False, repr=False)
class ValueRef(betterproto.Message):
    """ValueRef provides a unique identifier for a value"""

    store: str = betterproto.string_field(1)
    """The key/value store name"""

    key: str = betterproto.string_field(2)
    """The item's unique key within the store"""


@dataclass(eq=False, repr=False)
class Value(betterproto.Message):
    """Value provides a return value type"""

    ref: "ValueRef" = betterproto.message_field(1)
    """ValueRef of the key/value pair, which includes the store and key"""

    content: "betterproto_lib_google_protobuf.Struct" = betterproto.message_field(2)
    """The content (JSON object)"""


@dataclass(eq=False, repr=False)
class KvStoreGetValueRequest(betterproto.Message):
    ref: "ValueRef" = betterproto.message_field(1)
    """
    ValueRef of the key/value pair to get, which includes the store and key
    """


@dataclass(eq=False, repr=False)
class KvStoreGetValueResponse(betterproto.Message):
    value: "Value" = betterproto.message_field(1)
    """The retrieved value"""


@dataclass(eq=False, repr=False)
class KvStoreSetValueRequest(betterproto.Message):
    ref: "ValueRef" = betterproto.message_field(1)
    """
    ValueRef of the key/value pair to set, which includes the store and key
    """

    content: "betterproto_lib_google_protobuf.Struct" = betterproto.message_field(3)
    """The value content to store (JSON object)"""


@dataclass(eq=False, repr=False)
class KvStoreSetValueResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class KvStoreDeleteKeyRequest(betterproto.Message):
    ref: "ValueRef" = betterproto.message_field(1)
    """
    ValueRef of the key/value pair to delete, which includes the store and key
    """


@dataclass(eq=False, repr=False)
class KvStoreDeleteKeyResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class KvStoreScanKeysRequest(betterproto.Message):
    store: "Store" = betterproto.message_field(1)
    """The store to iterate over"""

    prefix: str = betterproto.string_field(2)
    """The prefix to filter keys by"""


@dataclass(eq=False, repr=False)
class KvStoreScanKeysResponse(betterproto.Message):
    key: str = betterproto.string_field(1)
    """The key of the key/value pair"""


class KvStoreStub(betterproto.ServiceStub):
    async def get_value(
        self,
        kv_store_get_value_request: "KvStoreGetValueRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "KvStoreGetValueResponse":
        return await self._unary_unary(
            "/nitric.proto.kvstore.v1.KvStore/GetValue",
            kv_store_get_value_request,
            KvStoreGetValueResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def set_value(
        self,
        kv_store_set_value_request: "KvStoreSetValueRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "KvStoreSetValueResponse":
        return await self._unary_unary(
            "/nitric.proto.kvstore.v1.KvStore/SetValue",
            kv_store_set_value_request,
            KvStoreSetValueResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def delete_key(
        self,
        kv_store_delete_key_request: "KvStoreDeleteKeyRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "KvStoreDeleteKeyResponse":
        return await self._unary_unary(
            "/nitric.proto.kvstore.v1.KvStore/DeleteKey",
            kv_store_delete_key_request,
            KvStoreDeleteKeyResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def scan_keys(
        self,
        kv_store_scan_keys_request: "KvStoreScanKeysRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> AsyncIterator["KvStoreScanKeysResponse"]:
        async for response in self._unary_stream(
            "/nitric.proto.kvstore.v1.KvStore/ScanKeys",
            kv_store_scan_keys_request,
            KvStoreScanKeysResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        ):
            yield response


class KvStoreBase(ServiceBase):
    async def get_value(
        self, kv_store_get_value_request: "KvStoreGetValueRequest"
    ) -> "KvStoreGetValueResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def set_value(
        self, kv_store_set_value_request: "KvStoreSetValueRequest"
    ) -> "KvStoreSetValueResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def delete_key(
        self, kv_store_delete_key_request: "KvStoreDeleteKeyRequest"
    ) -> "KvStoreDeleteKeyResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def scan_keys(
        self, kv_store_scan_keys_request: "KvStoreScanKeysRequest"
    ) -> AsyncIterator["KvStoreScanKeysResponse"]:
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)
        yield KvStoreScanKeysResponse()

    async def __rpc_get_value(
        self,
        stream: "grpclib.server.Stream[KvStoreGetValueRequest, KvStoreGetValueResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.get_value(request)
        await stream.send_message(response)

    async def __rpc_set_value(
        self,
        stream: "grpclib.server.Stream[KvStoreSetValueRequest, KvStoreSetValueResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.set_value(request)
        await stream.send_message(response)

    async def __rpc_delete_key(
        self,
        stream: "grpclib.server.Stream[KvStoreDeleteKeyRequest, KvStoreDeleteKeyResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.delete_key(request)
        await stream.send_message(response)

    async def __rpc_scan_keys(
        self,
        stream: "grpclib.server.Stream[KvStoreScanKeysRequest, KvStoreScanKeysResponse]",
    ) -> None:
        request = await stream.recv_message()
        await self._call_rpc_handler_server_stream(
            self.scan_keys,
            stream,
            request,
        )

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/nitric.proto.kvstore.v1.KvStore/GetValue": grpclib.const.Handler(
                self.__rpc_get_value,
                grpclib.const.Cardinality.UNARY_UNARY,
                KvStoreGetValueRequest,
                KvStoreGetValueResponse,
            ),
            "/nitric.proto.kvstore.v1.KvStore/SetValue": grpclib.const.Handler(
                self.__rpc_set_value,
                grpclib.const.Cardinality.UNARY_UNARY,
                KvStoreSetValueRequest,
                KvStoreSetValueResponse,
            ),
            "/nitric.proto.kvstore.v1.KvStore/DeleteKey": grpclib.const.Handler(
                self.__rpc_delete_key,
                grpclib.const.Cardinality.UNARY_UNARY,
                KvStoreDeleteKeyRequest,
                KvStoreDeleteKeyResponse,
            ),
            "/nitric.proto.kvstore.v1.KvStore/ScanKeys": grpclib.const.Handler(
                self.__rpc_scan_keys,
                grpclib.const.Cardinality.UNARY_STREAM,
                KvStoreScanKeysRequest,
                KvStoreScanKeysResponse,
            ),
        }



#
# Copyright (c) 2021 Nitric Technologies Pty Ltd.
#
# This file is part of Nitric Python 3 SDK.
# See https://github.com/nitrictech/python-sdk for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import asyncio
from typing import Generic, List, TypeVar

T = TypeVar("T")


class AsyncNotifierList(Generic[T]):
    """An async iterable that notifies when new items are added."""

    def __init__(self):
        """Create a new AsyncNotifierList."""
        self.items: List[T] = []  # type: ignore
        self.new_item_event: asyncio.Event = asyncio.Event()  # type: ignore

    async def add_item(self, item: T) -> None:
        """Add a new item to the list."""
        self.items.append(item)
        self.new_item_event.set()

    def __aiter__(self):
        return self

    async def __anext__(self):
        while not self.items:
            await self.new_item_event.wait()  # Wait for an item to be added
        item = self.items.pop(0)
        if not self.items:
            self.new_item_event.clear()  # Reset the event if there are no more items
        return item



#
# Copyright (c) 2021 Nitric Technologies Pty Ltd.
#
# This file is part of Nitric Python 3 SDK.
# See https://github.com/nitrictech/python-sdk for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: nitric/proto/resources/v1/resources.proto
# plugin: python-betterproto
# This file has been @generated

from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Dict,
    List,
    Optional,
)

import betterproto
import grpclib
from betterproto.grpc.grpclib_server import ServiceBase


if TYPE_CHECKING:
    import grpclib.server
    from betterproto.grpc.grpclib_client import MetadataLike
    from grpclib.metadata import Deadline


class ResourceType(betterproto.Enum):
    Api = 0
    Service = 1
    Bucket = 2
    Topic = 3
    Schedule = 4
    Subscription = 5
    KeyValueStore = 6
    Policy = 7
    Secret = 8
    BucketListener = 9
    Websocket = 10
    Http = 11
    ApiSecurityDefinition = 12
    Queue = 13


class Action(betterproto.Enum):
    BucketFileList = 0
    """Bucket Permissions: 0XX"""

    BucketFileGet = 1
    BucketFilePut = 2
    BucketFileDelete = 3
    TopicPublish = 200
    """Topic Permissions: 2XX"""

    KeyValueStoreRead = 300
    """KeyValue Store Permissions: 3XX"""

    KeyValueStoreWrite = 301
    KeyValueStoreDelete = 302
    SecretPut = 400
    """Secret Permissions: 4XX"""

    SecretAccess = 401
    WebsocketManage = 500
    """Websocket Permissions: 5XX"""

    QueueEnqueue = 600
    """Queue Permissions: 6XX"""

    QueueDequeue = 601


@dataclass(eq=False, repr=False)
class PolicyResource(betterproto.Message):
    principals: List["ResourceIdentifier"] = betterproto.message_field(1)
    actions: List["Action"] = betterproto.enum_field(2)
    resources: List["ResourceIdentifier"] = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class ResourceIdentifier(betterproto.Message):
    """Unique identifier for a resource within a nitric application."""

    type: "ResourceType" = betterproto.enum_field(1)
    name: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class ResourceDeclareRequest(betterproto.Message):
    id: "ResourceIdentifier" = betterproto.message_field(1)
    policy: "PolicyResource" = betterproto.message_field(10, group="config")
    bucket: "BucketResource" = betterproto.message_field(11, group="config")
    topic: "TopicResource" = betterproto.message_field(12, group="config")
    key_value_store: "KeyValueStoreResource" = betterproto.message_field(
        13, group="config"
    )
    secret: "SecretResource" = betterproto.message_field(14, group="config")
    api: "ApiResource" = betterproto.message_field(15, group="config")
    api_security_definition: "ApiSecurityDefinitionResource" = (
        betterproto.message_field(16, group="config")
    )
    queue: "QueueResource" = betterproto.message_field(17, group="config")


@dataclass(eq=False, repr=False)
class BucketResource(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class TopicResource(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class QueueResource(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class KeyValueStoreResource(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class SecretResource(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class ApiOpenIdConnectionDefinition(betterproto.Message):
    issuer: str = betterproto.string_field(1)
    audiences: List[str] = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class ApiSecurityDefinitionResource(betterproto.Message):
    api_name: str = betterproto.string_field(1)
    oidc: "ApiOpenIdConnectionDefinition" = betterproto.message_field(
        2, group="definition"
    )


@dataclass(eq=False, repr=False)
class ApiScopes(betterproto.Message):
    scopes: List[str] = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class ApiResource(betterproto.Message):
    security: Dict[str, "ApiScopes"] = betterproto.map_field(
        1, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )
    """
    root level security map for this api references
    ApiSecurityDefinitionResource(s)
    """


@dataclass(eq=False, repr=False)
class ResourceDeclareResponse(betterproto.Message):
    pass


class ResourcesStub(betterproto.ServiceStub):
    async def declare(
        self,
        resource_declare_request: "ResourceDeclareRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "ResourceDeclareResponse":
        return await self._unary_unary(
            "/nitric.proto.resources.v1.Resources/Declare",
            resource_declare_request,
            ResourceDeclareResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )


class ResourcesBase(ServiceBase):
    async def declare(
        self, resource_declare_request: "ResourceDeclareRequest"
    ) -> "ResourceDeclareResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_declare(
        self,
        stream: "grpclib.server.Stream[ResourceDeclareRequest, ResourceDeclareResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.declare(request)
        await stream.send_message(response)

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/nitric.proto.resources.v1.Resources/Declare": grpclib.const.Handler(
                self.__rpc_declare,
                grpclib.const.Cardinality.UNARY_UNARY,
                ResourceDeclareRequest,
                ResourceDeclareResponse,
            ),
        }



#
# Copyright (c) 2021 Nitric Technologies Pty Ltd.
#
# This file is part of Nitric Python 3 SDK.
# See https://github.com/nitrictech/python-sdk for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

import logging
from typing import Any, Callable, List, Literal

import betterproto
import grpclib
from grpclib import GRPCError
from grpclib.client import Channel

from nitric.application import Nitric
from nitric.bidi import AsyncNotifierList
from nitric.context import EventHandler, FunctionServer, MessageContext, MessageRequest
from nitric.exception import exception_from_grpc_error
from nitric.proto.resources.v1 import Action, ResourceDeclareRequest, ResourceIdentifier, ResourceType
from nitric.proto.topics.v1 import ClientMessage, TopicMessage
from nitric.proto.topics.v1 import MessageRequest as ProtoMessageRequest
from nitric.proto.topics.v1 import MessageResponse as ProtoMessageResponse
from nitric.proto.topics.v1 import RegistrationRequest, SubscriberStub
from nitric.proto.topics.v1 import TopicPublishRequest, TopicsStub
from nitric.resources.resource import SecureResource
from nitric.utils import dict_from_struct, new_default_channel, struct_from_dict

TopicPermission = Literal["publish"]


class TopicRef:
    """A reference to a deployed topic, used to interact with the topic at runtime."""

    _channel: Channel
    _topics_stub: TopicsStub
    name: str

    def __init__(self, name: str) -> None:
        """Construct a reference to a deployed Topic."""
        self._channel: Channel = new_default_channel()
        self._topics_stub = TopicsStub(channel=self._channel)
        self.name = name

    def __del__(self) -> None:
        # close the channel when this client is destroyed
        if self._channel is not None:
            self._channel.close()

    async def publish(
        self,
        message: dict[str, Any],
    ) -> None:
        """
        Publish a message to a topic, which can be subscribed to by other services.

        :param message: the event to publish
        :return: the published event, with the id added if one was auto-generated
        """
        if not isinstance(message, dict):
            raise ValueError("Message must be a dictionary")

        try:
            proto_message = TopicMessage(struct_payload=struct_from_dict(message))
            await self._topics_stub.publish(
                topic_publish_request=TopicPublishRequest(topic_name=self.name, message=proto_message)
            )
            return None
        except GRPCError as grpc_err:
            raise exception_from_grpc_error(grpc_err) from grpc_err


class Topic(SecureResource):
    """A topic resource, used for asynchronous messaging between functions."""

    name: str
    actions: List[Action]

    def __init__(self, name: str) -> None:
        """Declare a new topic resourced."""
        super().__init__(name)

    async def _register(self) -> None:
        try:
            await self._resources_stub.declare(
                resource_declare_request=ResourceDeclareRequest(id=self._to_resource_id())
            )
        except GRPCError as grpc_err:
            raise exception_from_grpc_error(grpc_err) from grpc_err

    def _to_resource_id(self) -> ResourceIdentifier:
        return ResourceIdentifier(name=self.name, type=ResourceType.Topic)

    def _perms_to_actions(self, *args: TopicPermission) -> List[Action]:
        _permMap: dict[TopicPermission, List[Action]] = {"publish": [Action.TopicPublish]}

        return [action for perm in args for action in _permMap[perm]]

    def allow(self, perm: TopicPermission, *args: TopicPermission) -> TopicRef:
        """Request the specified permissions to this resource."""
        str_args = [perm] + [str(permission) for permission in args]
        self._register_policy(*str_args)

        return TopicRef(self.name)

    def subscribe(self) -> Callable[[EventHandler], None]:
        """Create and return a subscription decorator for this topic."""

        def decorator(func: EventHandler) -> None:
            Subscriber(
                topic_name=self.name,
                handler=func,
            )

        return decorator


def _message_context_from_proto(msg: ProtoMessageRequest) -> MessageContext:
    return MessageContext(
        request=MessageRequest(
            data=dict_from_struct(msg.message.struct_payload),
            topic=msg.topic_name,
        )
    )


class Subscriber(FunctionServer):
    """A handler for topic messages."""

    _handler: EventHandler
    _registration_request: RegistrationRequest
    _responses: AsyncNotifierList[ClientMessage]

    def __init__(self, topic_name: str, handler: EventHandler):
        """Construct a new WebsocketHandler."""
        self._handler = handler
        self._responses = AsyncNotifierList()
        self._registration_request = RegistrationRequest(topic_name=topic_name)

        Nitric._register_worker(self)

    async def _message_request_iterator(self):
        # Register with the server
        yield ClientMessage(registration_request=self._registration_request)
        # wait for any responses for the server and send them
        async for response in self._responses:
            yield response

    async def start(self) -> None:
        """Register this subscriber and listen for messages."""
        channel = new_default_channel()
        server = SubscriberStub(channel=channel)

        try:
            async for server_msg in server.subscribe(self._message_request_iterator()):
                msg_type, _ = betterproto.which_one_of(server_msg, "content")

                if msg_type == "registration_response":
                    continue
                if msg_type == "message_request":
                    ctx = _message_context_from_proto(server_msg.message_request)

                    response: ClientMessage
                    try:
                        result = await self._handler(ctx)
                        ctx = result if result else ctx
                        response = ClientMessage(
                            id=server_msg.id, message_response=ProtoMessageResponse(success=ctx.res.success)
                        )
                    except Exception as e:  # pylint: disable=broad-except
                        logging.exception("An unhandled error occurred in a subscription event handler: %s", e)
                        response = ClientMessage(id=server_msg.id, message_response=ProtoMessageResponse(success=False))
                    await self._responses.add_item(response)
        except grpclib.exceptions.GRPCError as e:
            print(f"Stream terminated: {e.message}")
        except grpclib.exceptions.StreamTerminatedError:
            print("Stream from membrane closed.")
        finally:
            print("Closing client stream")
            channel.close()


def topic(name: str) -> Topic:
    """
    Create and register a topic.

    If a topic has already been registered with the same name, the original reference will be reused.
    """
    # type ignored because the create call are treated as protected.
    return Nitric._create_resource(Topic, name)  # type: ignore pylint: disable=protected-access



#
# Copyright (c) 2021 Nitric Technologies Pty Ltd.
#
# This file is part of Nitric Python 3 SDK.
# See https://github.com/nitrictech/python-sdk for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Nitric SDK Configuration Settings."""

import os


class Settings:
    """Nitric default and env settings helper class."""

    def __init__(self):
        """Construct a new Nitric settings helper object."""
        self.SERVICE_ADDRESS = os.environ.get("SERVICE_ADDRESS", "127.0.0.1:50051")


settings = Settings()



#
# Copyright (c) 2021 Nitric Technologies Pty Ltd.
#
# This file is part of Nitric Python 3 SDK.
# See https://github.com/nitrictech/python-sdk for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: nitric/proto/keyvalue/v1/keyvalue.proto
# plugin: python-betterproto
# This file has been @generated

from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Dict,
    Optional,
)

import betterproto
import betterproto.lib.google.protobuf as betterproto_lib_google_protobuf
import grpclib
from betterproto.grpc.grpclib_server import ServiceBase

from .... import proto as ___proto__


if TYPE_CHECKING:
    import grpclib.server
    from betterproto.grpc.grpclib_client import MetadataLike
    from grpclib.metadata import Deadline


@dataclass(eq=False, repr=False)
class Store(betterproto.Message):
    """Provides a Key/Value Store"""

    name: str = betterproto.string_field(1)
    """The store name"""


@dataclass(eq=False, repr=False)
class ValueRef(betterproto.Message):
    """ValueRef provides a unique identifier for a value"""

    store: str = betterproto.string_field(1)
    """The key/value store name"""

    key: str = betterproto.string_field(2)
    """The item's unique key within the store"""


@dataclass(eq=False, repr=False)
class Value(betterproto.Message):
    """Value provides a return value type"""

    ref: "___proto__.KeyValueV1ValueRef" = betterproto.message_field(1)
    """ValueRef of the key/value pair, which includes the store and key"""

    content: "betterproto_lib_google_protobuf.Struct" = betterproto.message_field(2)
    """The content (JSON object)"""


@dataclass(eq=False, repr=False)
class KeyValueGetRequest(betterproto.Message):
    ref: "___proto__.KeyValueV1ValueRef" = betterproto.message_field(1)
    """
    ValueRef of the key/value pair to get, which includes the store and key
    """


@dataclass(eq=False, repr=False)
class KeyValueGetResponse(betterproto.Message):
    value: "___proto__.KeyValueV1Value" = betterproto.message_field(1)
    """The retrieved value"""


@dataclass(eq=False, repr=False)
class KeyValueSetRequest(betterproto.Message):
    ref: "___proto__.KeyValueV1ValueRef" = betterproto.message_field(1)
    """
    ValueRef of the key/value pair to set, which includes the store and key
    """

    content: "betterproto_lib_google_protobuf.Struct" = betterproto.message_field(3)
    """The value content to store (JSON object)"""


@dataclass(eq=False, repr=False)
class KeyValueSetResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class KeyValueDeleteRequest(betterproto.Message):
    ref: "___proto__.KeyValueV1ValueRef" = betterproto.message_field(1)
    """
    ValueRef of the key/value pair to delete, which includes the store and key
    """


@dataclass(eq=False, repr=False)
class KeyValueDeleteResponse(betterproto.Message):
    pass


class KeyValueStub(betterproto.ServiceStub):
    async def get(
        self,
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "___proto__.KeyValueV1KeyValueGetResponse":
        return await self._unary_unary(
            "/nitric.proto.KeyValue.v1.KeyValue/Get",
            proto_key_value_v1_key_value_get_request,
            ___proto__.KeyValueV1KeyValueGetResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def set(
        self,
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "___proto__.KeyValueV1KeyValueSetResponse":
        return await self._unary_unary(
            "/nitric.proto.KeyValue.v1.KeyValue/Set",
            proto_key_value_v1_key_value_set_request,
            ___proto__.KeyValueV1KeyValueSetResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def delete(
        self,
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "___proto__.KeyValueV1KeyValueDeleteResponse":
        return await self._unary_unary(
            "/nitric.proto.KeyValue.v1.KeyValue/Delete",
            proto_key_value_v1_key_value_delete_request,
            ___proto__.KeyValueV1KeyValueDeleteResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )


class KeyValueBase(ServiceBase):
    async def get(self) -> "___proto__.KeyValueV1KeyValueGetResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def set(self) -> "___proto__.KeyValueV1KeyValueSetResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def delete(self) -> "___proto__.KeyValueV1KeyValueDeleteResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_get(
        self,
        stream: "grpclib.server.Stream[___proto__.KeyValueV1KeyValueGetRequest, ___proto__.KeyValueV1KeyValueGetResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.get(request)
        await stream.send_message(response)

    async def __rpc_set(
        self,
        stream: "grpclib.server.Stream[___proto__.KeyValueV1KeyValueSetRequest, ___proto__.KeyValueV1KeyValueSetResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.set(request)
        await stream.send_message(response)

    async def __rpc_delete(
        self,
        stream: "grpclib.server.Stream[___proto__.KeyValueV1KeyValueDeleteRequest, ___proto__.KeyValueV1KeyValueDeleteResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.delete(request)
        await stream.send_message(response)

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/nitric.proto.KeyValue.v1.KeyValue/Get": grpclib.const.Handler(
                self.__rpc_get,
                grpclib.const.Cardinality.UNARY_UNARY,
                ___proto__.KeyValueV1KeyValueGetRequest,
                ___proto__.KeyValueV1KeyValueGetResponse,
            ),
            "/nitric.proto.KeyValue.v1.KeyValue/Set": grpclib.const.Handler(
                self.__rpc_set,
                grpclib.const.Cardinality.UNARY_UNARY,
                ___proto__.KeyValueV1KeyValueSetRequest,
                ___proto__.KeyValueV1KeyValueSetResponse,
            ),
            "/nitric.proto.KeyValue.v1.KeyValue/Delete": grpclib.const.Handler(
                self.__rpc_delete,
                grpclib.const.Cardinality.UNARY_UNARY,
                ___proto__.KeyValueV1KeyValueDeleteRequest,
                ___proto__.KeyValueV1KeyValueDeleteResponse,
            ),
        }



#
# Copyright (c) 2021 Nitric Technologies Pty Ltd.
#
# This file is part of Nitric Python 3 SDK.
# See https://github.com/nitrictech/python-sdk for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

import logging
from datetime import timedelta
from enum import Enum
from typing import Callable, List

import betterproto
import grpclib.exceptions

from nitric.application import Nitric
from nitric.bidi import AsyncNotifierList
from nitric.context import FunctionServer, IntervalContext, IntervalHandler
from nitric.proto.schedules.v1 import (
    ClientMessage,
    IntervalResponse,
    RegistrationRequest,
    ScheduleCron,
    ScheduleEvery,
    SchedulesStub,
)
from nitric.utils import new_default_channel


class ScheduleServer(FunctionServer):
    """A schedule for running functions on a cadence."""

    description: str

    handler: IntervalHandler
    _registration_request: RegistrationRequest
    _responses: AsyncNotifierList[ClientMessage]

    def __init__(self, description: str):
        """Create a schedule for running functions on a cadence."""
        self.description = description
        self._responses = AsyncNotifierList()

    def every(self, rate_description: str, handler: IntervalHandler) -> None:
        """
        Register a function to be run at the specified rate.

        E.g. every("3 hours")
        """
        self._registration_request = RegistrationRequest(
            schedule_name=self.description,
            every=ScheduleEvery(rate=rate_description.lower()),
        )

        self.handler = handler

        Nitric._register_worker(self)  # type: ignore pylint: disable=protected-access

    def cron(self, cron_expression: str, handler: IntervalHandler) -> None:
        """
        Register a function to be run at the specified cron schedule.

        E.g. cron("3 * * * *")
        """
        self._registration_request = RegistrationRequest(
            schedule_name=self.description,
            cron=ScheduleCron(expression=cron_expression),
        )

        self.handler = handler

        Nitric._register_worker(self)  # type: ignore pylint: disable=protected-access

    async def _schedule_request_iterator(self):
        # Register with the server
        yield ClientMessage(registration_request=self._registration_request)
        # wait for any responses for the server and send them
        async for response in self._responses:
            yield response

    async def start(self) -> None:
        """Register this schedule and start listening for requests."""
        channel = new_default_channel()
        schedules_stub = SchedulesStub(channel=channel)

        try:
            async for server_msg in schedules_stub.schedule(self._schedule_request_iterator()):
                msg_type, _ = betterproto.which_one_of(server_msg, "content")

                if msg_type == "registration_response":
                    continue
                if msg_type == "interval_request":
                    ctx = IntervalContext(server_msg)
                    try:
                        await self.handler(ctx)
                    except Exception as e:  # pylint: disable=broad-except
                        logging.exception("An unhandled error occurred in a scheduled function: %s", e)
                    resp = IntervalResponse()
                    await self._responses.add_item(ClientMessage(id=server_msg.id, interval_response=resp))
        except grpclib.exceptions.GRPCError as e:
            print(f"Stream terminated: {e.message}")
        except grpclib.exceptions.StreamTerminatedError:
            print("Stream from membrane closed.")
        finally:
            print("Closing client stream")
            channel.close()


class Frequency(Enum):
    """Valid schedule frequencies."""

    MINUTES = "minutes"
    HOURS = "hours"
    DAYS = "days"

    @staticmethod
    def from_str(value: str) -> Frequency:
        """Convert a string frequency value to Frequency."""
        try:
            return Frequency[value.strip().upper()]
        except Exception:
            raise ValueError(f"{value} is not a valid frequency")

    @staticmethod
    def as_str_list() -> List[str]:
        """Return all frequency values as a list of strings."""
        return [str(frequency.value) for frequency in Frequency]

    def as_time(self, rate: int) -> timedelta:
        """Convert the rate to minutes based on the frequency."""
        if self == Frequency.MINUTES:
            return timedelta(minutes=rate)
        elif self == Frequency.HOURS:
            return timedelta(hours=rate)
        elif self == Frequency.DAYS:
            return timedelta(days=rate)
        else:
            raise ValueError(f"{self} is not a valid frequency")


class Schedule:
    """A raw schedule to be deployed and assigned a rate or cron interval."""

    def __init__(self, description: str):
        """Create a new schedule resource."""
        self.description = description

    def every(self, every: str) -> Callable[[IntervalHandler], ScheduleServer]:
        """
        Set the schedule interval.

        e.g. every('3 days').
        """

        def decorator(func: IntervalHandler) -> ScheduleServer:
            r = ScheduleServer(self.description)
            r.every(every, func)
            return r

        return decorator

    def cron(self, cron: str) -> Callable[[IntervalHandler], ScheduleServer]:
        """
        Set the schedule interval.

        e.g. cron('3 * * * *').
        """

        def decorator(func: IntervalHandler) -> ScheduleServer:
            r = ScheduleServer(self.description)
            r.cron(cron, func)
            return r

        return decorator


def schedule(description: str) -> Schedule:
    """Return a schedule."""
    return Schedule(description=description)



#
# Copyright (c) 2021 Nitric Technologies Pty Ltd.
#
# This file is part of Nitric Python 3 SDK.
# See https://github.com/nitrictech/python-sdk for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

import asyncio
from abc import ABC, abstractmethod
from asyncio import Task
from typing import Any, List, Optional, Sequence, Type, TypeVar

from grpclib import GRPCError

from nitric.exception import NitricResourceException, exception_from_grpc_error
from nitric.proto.resources.v1 import (
    Action,
    PolicyResource,
    ResourceDeclareRequest,
    ResourceIdentifier,
    ResourcesStub,
    ResourceType,
)
from nitric.utils import new_default_channel

T = TypeVar("T", bound="Resource")


class Resource(ABC):
    """A base resource class with common functionality."""

    name: str

    def __init__(self, name: str):
        """Construct a new resource."""
        self.name = name
        self._reg: Optional[Task[Any]] = None  # type: ignore
        self._channel = new_default_channel()
        self._resources_stub = ResourcesStub(channel=self._channel)

    @abstractmethod
    async def _register(self) -> None:
        pass

    @classmethod
    def make(cls: Type[T], name: str, *args: Sequence[Any], **kwargs: dict[str, Any]) -> T:
        """
        Create and register the resource.

        The registration process for resources async, so this method should be used instead of __init__.
        """
        r = cls(name, *args, **kwargs)  # type: ignore
        try:
            loop = asyncio.get_running_loop()
            r._reg = loop.create_task(r._register())
        except RuntimeError:
            loop = asyncio.get_event_loop()
            loop.run_until_complete(r._register())

        return r


class SecureResource(Resource):
    """A secure base resource class."""

    @abstractmethod
    def _to_resource_id(self) -> ResourceIdentifier:
        pass

    @abstractmethod
    def _perms_to_actions(self, *args: Any) -> List[Action]:
        pass

    async def _register_policy_async(self, *args: str) -> None:
        # if self._reg is not None:
        #     await asyncio.wait({self._reg})

        policy = PolicyResource(
            principals=[ResourceIdentifier(type=ResourceType.Service)],
            actions=self._perms_to_actions(*args),
            resources=[self._to_resource_id()],
        )
        try:
            await self._resources_stub.declare(
                resource_declare_request=ResourceDeclareRequest(
                    id=ResourceIdentifier(type=ResourceType.Policy), policy=policy
                )
            )
        except GRPCError as grpc_err:
            raise exception_from_grpc_error(grpc_err) from grpc_err

    def _register_policy(self, *args: str) -> None:
        try:
            loop = asyncio.get_event_loop()
            loop.run_until_complete(self._register_policy_async(*args))
        except RuntimeError:
            raise NitricResourceException(
                "Nitric resources cannot be declared at runtime e.g. within the scope of a runtime function. \
                    Move resource declarations to the top level of scripts so that they can be safely provisioned"
            ) from None



#
# Copyright (c) 2021 Nitric Technologies Pty Ltd.
#
# This file is part of Nitric Python 3 SDK.
# See https://github.com/nitrictech/python-sdk for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: nitric/proto/websockets/v1/websockets.proto
# plugin: python-betterproto
# This file has been @generated

from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    AsyncIterable,
    AsyncIterator,
    Dict,
    Iterable,
    List,
    Optional,
    Union,
)

import betterproto
import grpclib
from betterproto.grpc.grpclib_server import ServiceBase


if TYPE_CHECKING:
    import grpclib.server
    from betterproto.grpc.grpclib_client import MetadataLike
    from grpclib.metadata import Deadline


class WebsocketEventType(betterproto.Enum):
    Connect = 0
    """Specialised Event for handling new client connections"""

    Disconnect = 1
    """Specialised Event for handling existing client connections"""

    Message = 2
    """All other types of events are messages"""


@dataclass(eq=False, repr=False)
class WebsocketDetailsRequest(betterproto.Message):
    socket_name: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class WebsocketDetailsResponse(betterproto.Message):
    url: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class WebsocketSendRequest(betterproto.Message):
    socket_name: str = betterproto.string_field(1)
    """The nitric name of the socket to send on"""

    connection_id: str = betterproto.string_field(2)
    """The connection ID of the client to send to"""

    data: bytes = betterproto.bytes_field(3)
    """The data to send to the socket"""


@dataclass(eq=False, repr=False)
class WebsocketSendResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class WebsocketCloseConnectionRequest(betterproto.Message):
    socket_name: str = betterproto.string_field(1)
    """The nitric name of the socket to send on"""

    connection_id: str = betterproto.string_field(2)
    """The connection ID of the client to send to"""


@dataclass(eq=False, repr=False)
class WebsocketCloseConnectionResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class ClientMessage(betterproto.Message):
    """ClientMessages are sent from the service to the nitric server"""

    id: str = betterproto.string_field(1)
    """Globally unique id to pair requests/responses"""

    registration_request: "RegistrationRequest" = betterproto.message_field(
        2, group="content"
    )
    """
    Client initialisation request A worker will not be eligible for triggers
    until it has identified itself
    """

    websocket_event_response: "WebsocketEventResponse" = betterproto.message_field(
        3, group="content"
    )
    """Client responding with result of a trigger"""


@dataclass(eq=False, repr=False)
class RegistrationResponse(betterproto.Message):
    """Placeholder message"""

    pass


@dataclass(eq=False, repr=False)
class RegistrationRequest(betterproto.Message):
    socket_name: str = betterproto.string_field(1)
    """The nitric name of the socket that this worker listens on"""

    event_type: "WebsocketEventType" = betterproto.enum_field(2)
    """The type of event that this worker handles"""


@dataclass(eq=False, repr=False)
class WebsocketEventRequest(betterproto.Message):
    socket_name: str = betterproto.string_field(1)
    """The nitric name of the socket that this worker listens on"""

    connection_id: str = betterproto.string_field(2)
    """The connection this trigger came from"""

    connection: "WebsocketConnectionEvent" = betterproto.message_field(
        10, group="websocket_event"
    )
    disconnection: "WebsocketDisconnectionEvent" = betterproto.message_field(
        11, group="websocket_event"
    )
    message: "WebsocketMessageEvent" = betterproto.message_field(
        12, group="websocket_event"
    )


@dataclass(eq=False, repr=False)
class QueryValue(betterproto.Message):
    value: List[str] = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class ServerMessage(betterproto.Message):
    """ServerMessages are sent from the nitric server to the service"""

    id: str = betterproto.string_field(1)
    """Server message ID, used to pair requests/responses"""

    registration_response: "RegistrationResponse" = betterproto.message_field(
        2, group="content"
    )
    """
    Server responding with client configuration details to an InitRequest
    """

    websocket_event_request: "WebsocketEventRequest" = betterproto.message_field(
        3, group="content"
    )
    """Server requesting client to process an event"""


@dataclass(eq=False, repr=False)
class WebsocketEventResponse(betterproto.Message):
    connection_response: "WebsocketConnectionResponse" = betterproto.message_field(
        10, group="websocket_response"
    )


@dataclass(eq=False, repr=False)
class WebsocketConnectionEvent(betterproto.Message):
    query_params: Dict[str, "QueryValue"] = betterproto.map_field(
        1, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )
    """The query params available in the connection request"""


@dataclass(eq=False, repr=False)
class WebsocketConnectionResponse(betterproto.Message):
    reject: bool = betterproto.bool_field(1)


@dataclass(eq=False, repr=False)
class WebsocketDisconnectionEvent(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class WebsocketMessageEvent(betterproto.Message):
    body: bytes = betterproto.bytes_field(1)
    """Data available on"""


class WebsocketStub(betterproto.ServiceStub):
    async def send_message(
        self,
        websocket_send_request: "WebsocketSendRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "WebsocketSendResponse":
        return await self._unary_unary(
            "/nitric.proto.websockets.v1.Websocket/SendMessage",
            websocket_send_request,
            WebsocketSendResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def close_connection(
        self,
        websocket_close_connection_request: "WebsocketCloseConnectionRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "WebsocketCloseConnectionResponse":
        return await self._unary_unary(
            "/nitric.proto.websockets.v1.Websocket/CloseConnection",
            websocket_close_connection_request,
            WebsocketCloseConnectionResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def socket_details(
        self,
        websocket_details_request: "WebsocketDetailsRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "WebsocketDetailsResponse":
        return await self._unary_unary(
            "/nitric.proto.websockets.v1.Websocket/SocketDetails",
            websocket_details_request,
            WebsocketDetailsResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )


class WebsocketHandlerStub(betterproto.ServiceStub):
    async def handle_events(
        self,
        client_message_iterator: Union[
            AsyncIterable["ClientMessage"], Iterable["ClientMessage"]
        ],
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> AsyncIterator["ServerMessage"]:
        async for response in self._stream_stream(
            "/nitric.proto.websockets.v1.WebsocketHandler/HandleEvents",
            client_message_iterator,
            ClientMessage,
            ServerMessage,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        ):
            yield response


class WebsocketBase(ServiceBase):
    async def send_message(
        self, websocket_send_request: "WebsocketSendRequest"
    ) -> "WebsocketSendResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def close_connection(
        self, websocket_close_connection_request: "WebsocketCloseConnectionRequest"
    ) -> "WebsocketCloseConnectionResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def socket_details(
        self, websocket_details_request: "WebsocketDetailsRequest"
    ) -> "WebsocketDetailsResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_send_message(
        self,
        stream: "grpclib.server.Stream[WebsocketSendRequest, WebsocketSendResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.send_message(request)
        await stream.send_message(response)

    async def __rpc_close_connection(
        self,
        stream: "grpclib.server.Stream[WebsocketCloseConnectionRequest, WebsocketCloseConnectionResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.close_connection(request)
        await stream.send_message(response)

    async def __rpc_socket_details(
        self,
        stream: "grpclib.server.Stream[WebsocketDetailsRequest, WebsocketDetailsResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.socket_details(request)
        await stream.send_message(response)

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/nitric.proto.websockets.v1.Websocket/SendMessage": grpclib.const.Handler(
                self.__rpc_send_message,
                grpclib.const.Cardinality.UNARY_UNARY,
                WebsocketSendRequest,
                WebsocketSendResponse,
            ),
            "/nitric.proto.websockets.v1.Websocket/CloseConnection": grpclib.const.Handler(
                self.__rpc_close_connection,
                grpclib.const.Cardinality.UNARY_UNARY,
                WebsocketCloseConnectionRequest,
                WebsocketCloseConnectionResponse,
            ),
            "/nitric.proto.websockets.v1.Websocket/SocketDetails": grpclib.const.Handler(
                self.__rpc_socket_details,
                grpclib.const.Cardinality.UNARY_UNARY,
                WebsocketDetailsRequest,
                WebsocketDetailsResponse,
            ),
        }


class WebsocketHandlerBase(ServiceBase):
    async def handle_events(
        self, client_message_iterator: AsyncIterator["ClientMessage"]
    ) -> AsyncIterator["ServerMessage"]:
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)
        yield ServerMessage()

    async def __rpc_handle_events(
        self, stream: "grpclib.server.Stream[ClientMessage, ServerMessage]"
    ) -> None:
        request = stream.__aiter__()
        await self._call_rpc_handler_server_stream(
            self.handle_events,
            stream,
            request,
        )

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/nitric.proto.websockets.v1.WebsocketHandler/HandleEvents": grpclib.const.Handler(
                self.__rpc_handle_events,
                grpclib.const.Cardinality.STREAM_STREAM,
                ClientMessage,
                ServerMessage,
            ),
        }



#
# Copyright (c) 2021 Nitric Technologies Pty Ltd.
#
# This file is part of Nitric Python 3 SDK.
# See https://github.com/nitrictech/python-sdk for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

from typing import Any, List, Literal, AsyncIterator, Optional

from grpclib import GRPCError
from grpclib.client import Channel

from nitric.application import Nitric
from nitric.exception import exception_from_grpc_error
from nitric.proto.kvstore.v1 import (
    Store,
    KvStoreScanKeysRequest,
    KvStoreDeleteKeyRequest,
    KvStoreGetValueRequest,
    KvStoreSetValueRequest,
    KvStoreStub,
    ValueRef,
)
from nitric.proto.resources.v1 import (
    Action,
    KeyValueStoreResource,
    ResourceDeclareRequest,
    ResourceIdentifier,
    ResourceType,
)
from nitric.resources.resource import SecureResource
from nitric.utils import dict_from_struct, new_default_channel, struct_from_dict


class KeyValueStoreRef:
    """A reference to a deployed key value store, used to interact with the key value store at runtime."""

    _kv_stub: KvStoreStub
    _channel: Channel
    name: str

    def __init__(self, name: str):
        """Construct a reference to a deployed key value store."""
        self._channel: Channel = new_default_channel()
        self._kv_stub = KvStoreStub(channel=self._channel)
        self.name = name

    def __del__(self):
        # close the channel when this client is destroyed
        if self._channel is not None:
            self._channel.close()

    async def set(self, key: str, value: dict[str, Any]) -> None:
        """Set a key and value in the key value store."""
        ref = ValueRef(store=self.name, key=key)

        req = KvStoreSetValueRequest(ref=ref, content=struct_from_dict(value))

        try:
            await self._kv_stub.set_value(kv_store_set_value_request=req)
        except GRPCError as grpc_err:
            raise exception_from_grpc_error(grpc_err) from grpc_err

    async def get(self, key: str) -> dict[str, Any]:
        """Return a value from the key value store."""
        ref = ValueRef(store=self.name, key=key)

        req = KvStoreGetValueRequest(ref=ref)

        try:
            resp = await self._kv_stub.get_value(kv_store_get_value_request=req)

            return dict_from_struct(resp.value.content)
        except GRPCError as grpc_err:
            raise exception_from_grpc_error(grpc_err) from grpc_err

    async def keys(self, prefix: Optional[str] = "") -> AsyncIterator[str]:
        """Return a list of keys from the key value store."""
        if prefix is None:
            prefix = ""

        req = KvStoreScanKeysRequest(
            store=Store(name=self.name),
            prefix=prefix,
        )

        try:
            response_iterator = self._kv_stub.scan_keys(kv_store_scan_keys_request=req)
            async for item in response_iterator:
                yield item.key
        except GRPCError as grpc_err:
            raise exception_from_grpc_error(grpc_err) from grpc_err

        return

    async def delete(self, key: str) -> None:
        """Delete a key from the key value store."""
        ref = ValueRef(store=self.name, key=key)

        req = KvStoreDeleteKeyRequest(ref=ref)

        try:
            await self._kv_stub.delete_key(kv_store_delete_key_request=req)
        except GRPCError as grpc_err:
            raise exception_from_grpc_error(grpc_err) from grpc_err


KVPermission = Literal["get", "set", "delete"]


class KeyValueStore(SecureResource):
    """A key value store resource."""

    async def _register(self) -> None:
        try:
            await self._resources_stub.declare(
                resource_declare_request=ResourceDeclareRequest(
                    id=self._to_resource_id(), key_value_store=KeyValueStoreResource()
                )
            )
        except GRPCError as grpc_err:
            raise exception_from_grpc_error(grpc_err) from grpc_err

    def _to_resource_id(self) -> ResourceIdentifier:
        return ResourceIdentifier(name=self.name, type=ResourceType.KeyValueStore)

    def _perms_to_actions(self, *args: KVPermission) -> List[Action]:
        permission_actions_map: dict[KVPermission, List[Action]] = {
            "get": [Action.KeyValueStoreRead],
            "set": [Action.KeyValueStoreWrite],
            "delete": [Action.KeyValueStoreDelete],
        }

        return [action for perm in args for action in permission_actions_map[perm]]

    def allow(self, perm: KVPermission, *args: KVPermission) -> KeyValueStoreRef:
        """Request the required permissions for this collection."""
        # Ensure registration of the resource is complete before requesting permissions.
        str_args = [str(perm)] + [str(permission) for permission in args]
        self._register_policy(*str_args)

        return KeyValueStoreRef(self.name)


def kv(name: str) -> KeyValueStore:
    """
    Create and register a key value store.

    If a key value store has already been registered with the same name, the original reference will be reused.
    """
    return Nitric._create_resource(KeyValueStore, name)  # type: ignore pylint: disable=protected-access
